﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;

namespace DragonChat
{
    public class ALSort : IComparer
    {
        private CaseInsensitiveComparer ObjectCompare;

        public ALSort()
        {
            ObjectCompare = new CaseInsensitiveComparer();
        }

        public int Compare(object x, object y)
        {
            int CompareResult;

           AccessEntry userX, userY;

           userX = (AccessEntry)((ListViewItem)x).Tag;
           userY = (AccessEntry)((ListViewItem)y).Tag;

           CompareResult = ObjectCompare.Compare(userX.Order, userY.Order);
           if (CompareResult == 0)
                return ObjectCompare.Compare(userX.SetbyNick, userY.SetbyNick);
            else
                return CompareResult;
        }
    }

    public class AccessCollection : CollectionBase
    {

        public void Add(AccessEntry user)
        {
            base.List.Add(user);
        }

        public void Remove(AccessEntry user)
        {
            base.List.Remove(user);
        }

        public bool RemoveEntry(string entry, string level)
        {
            foreach (AccessEntry e in base.List)
            {
                if (entry == e.Entry)
                {
                    if (level == e.Level)
                    {
                        base.List.Remove(e);
                        return true;
                    }
                }
            }
            return false;
        }

        public AccessEntry[] GetList()
        {
            if (base.List.Count == 0)
                return null;
            AccessEntry[] Entries = new AccessEntry[base.List.Count];
            int i = 0;
            foreach (AccessEntry e in base.List)
            {
                Entries[i] = e;
                i++;
            }
            return Entries;
        }
    }

    public class AccessEntry
    {
        string _entry = string.Empty;
        string _setby = string.Empty;
        string _tag = string.Empty;
        string _level = string.Empty;
        string _duration = string.Empty;

        public AccessEntry(string entry, string setby, string level, string duration, string tag)
        {
            this._entry = entry;
            this._duration = duration;
            this._level = level;
            this._setby = setby;
            this._tag = tag;
        }

        public string Entry
        {
            get
            {
                return _entry;
            }
        }

        public string Level
        {
            get
            {
                return _level;
            }
        }

        public string Duration
        {
            get
            {
                if (_duration == "0")
                    return "Permanent";
                else
                {
                    StringBuilder sb = new StringBuilder();
                    try
                    {
                        TimeSpan ts = TimeSpan.FromMinutes(Convert.ToDouble(_duration));
                        if (ts.Days > 0)
                        {
                            sb.Append(ts.Days);
                            sb.Append("day");
                        }
                        if (ts.Hours > 0)
                        {
                            sb.Append(ts.Hours);
                            sb.Append("hrs");
                        }
                        if (ts.Minutes > 0)
                        {
                            sb.Append(ts.Minutes);
                            sb.Append("min");
                        }
                    }
                    catch (Exception)
                    {
                        return _duration;
                    }
                    return sb.ToString();
                }
            }
        }

        public string Tag
        {
            get
            {
                return _tag;
            }
        }

        public int Order
        {
            get
            {
                if (_level == "OWNER")
                    return 0;
                if (_level == "HOST")
                    return 1;
                if (_level == "VOICE")
                    return 2;
                if (_level == "GRANT")
                    return 3;
                if (_level == "DENY")
                    return 4;
                return 5;
            }
        }

        public string SetbyNick
        {
            get
            {
                try
                {
                    return _setby.Substring(0, _setby.LastIndexOf('!'));
                }
                catch (Exception)
                {
                    return "Unknown User";
                }
            }
        }

        public string SetbyAddress
        {
            get
            {
                try
                {
                    return _setby.Substring(_setby.LastIndexOf('!') + 1);
                }
                catch
                {
                    return "Unknown User";
                }
            }
        }
    }
}
