using System;
using System.Collections.Generic;
using System.Text;

namespace DragonChat
{
    public class User
    {
        private string _nickname = "";

        private string _utf8 = "";

        private string _address = "";

        private bool _staff = false;
        private bool _guest = false;

        private bool _super = false;
        private bool _owner = false;
        private bool _host = false;
        private bool _helpop = false;

        private bool _ignore = false;

        private bool _away = false;

        private bool _voice = false;

        private bool _moderated = false;

        private ProfileType _profile = ProfileType.NoProfile;

        public User(string nick, string pdata)
        {
            Nickname = nick;
            if (_guest == false)
                SetProfile(pdata);
        }

        public User(string nick, string addr, string pdata)
        {
            Nickname = nick;
            Address = addr;
            if (_guest == false)
                SetProfile(pdata);
        }

        public int Order
        {
            get
            {
                if (_staff)
                    return 1;
                if (_super)
                {
                    if (!_guest)
                        return 2;
                    else
                        return 3;
                }
                if (_owner)
                {
                    if (!_guest)
                        return 4;
                    else
                        return 5;
                }
                if (_host)
                {
                    if (!_guest)
                        return 6;
                    else
                        return 7;
                }
                if (_helpop)
                {
                    if (!_guest)
                        return 8;
                    else
                        return 9;
                }
                if (_moderated)
                {
                    if (_voice)
                    {
                        if (!_guest)
                            return 10;
                        else
                            return 11;
                    }
                }
                if (!_guest)
                    return 12;
                else
                    return 13;
            }
        }

        private void SetProfile(string pdata)
        {
            if (pdata.Substring(0, 1) == "A")
                Away = true;
            if (pdata.Substring(3, 1) == "V")
                Voice = true;
            switch (pdata.Substring(1, 2))
            {
                case "MN":
                    _profile = ProfileType.MaleNoPic;
                    break;
                case "MP":
                    _profile = ProfileType.MaleWithPic;
                    break;
                case "FN":
                    _profile = ProfileType.FemaleNoPic;
                    break;
                case "FP":
                    _profile = ProfileType.FemaleWithPic;
                    break;
                case "UN":
                    _profile = ProfileType.UnknownNoPic;
                    break;
                case "UP":
                    _profile = ProfileType.UnknownWithPic;
                    break;
                default:
                    _profile = ProfileType.NoProfile;
                    break;
            }
        }

        public string Nickname
        {
            get
            {
                return _nickname;
            }
            set
            {
                _utf8 = ToUTF8(value);
                _nickname = value;
                foreach (Char c in value.ToCharArray())
                {
                    bool end = false;
                    switch (c)
                    {
                        case (char)39:
                            _super = true;
                            _nickname = value.Substring(1);
                            break;
                        case '.':
                            _owner = true;
                            _nickname = value.Substring(1);
                            break;
                        case '@':
                            _host = true;
                            _nickname = value.Substring(1);
                            break;
                        case '%':
                            _helpop = true;
                            _nickname = value.Substring(1);
                            break;
                        default:
                            end = true;
                            break;
                    }
                    if (end)
                        break;
                }
                if (_nickname.Substring(0, 1) == "^")
                    _staff = true;
                if (_nickname.Substring(0, 1) == ">")
                    _guest = true;
            }
        }

        public ProfileType Profile
        {
            get
            {
                return _profile;
            }
            set
            {
                _profile = value;
            }
        }

        public enum ProfileType
        {
            MaleNoPic = 5,
            MaleWithPic = 6,
            FemaleNoPic = 7,
            FemaleWithPic = 8,
            UnknownNoPic = 9,
            UnknownWithPic = 10,
            NoProfile = 11
        }

        public string Address
        {
            get
            {
                return _address;
            }
            set
            {
                _address = value;
            }
        }

        public bool Guest
        {
            get
            {
                return _guest;
            }
        }

        public bool Superowner
        {
            get
            {
                return _super;
            }
            set
            {
                _super = value;
            }
        }

        public bool Owner
        {
            get
            {
                return _owner;
            }
            set
            {
                _owner = value;
            }
        }

        public bool Host
        {
            get
            {
                return _host;
            }
            set
            {
                _host = value;
            }
        }

        public bool HelpOp
        {
            get
            {
                return _helpop;
            }
            set
            {
                _helpop = value;
            }
        }

        public bool Staff
        {
            get
            {
                return _staff;
            }
        }

        public bool Away
        {
            get
            {
                return _away;
            }
            set
            {
                _away = value;
            }
        }

        public bool Ignore
        {
            get
            {
                return _ignore;
            }
            set
            {
                _ignore = value;
            }
        }

        public bool Moderated
        {
            get
            {
                return _moderated;
            }
            set
            {
                _moderated = value;
            }
        }

        public bool Voice
        {
            get
            {
                return _voice;
            }
            set
            {
                _voice = value;
            }
        }

        public int ImageIndex
        {
            get
            {
                if (_ignore)
                    return 14;
                if (_away)
                    return 12;
                if (_staff)
                    return 0;
                if (_super)
                    return 1;
                if (_owner)
                    return 2;
                if (_host)
                    return 3;
                if (_helpop)
                    return 4;
                if (_moderated)
                    if (!_voice)
                        return 13;
                return (int)_profile;
            }
        }

        private string ToUTF8(string text)
        {
            byte[] b = Encoding.Unicode.GetBytes(text);
            b = Encoding.Convert(Encoding.Unicode, Encoding.UTF8, b);
            return Encoding.Default.GetString(b);
        }

        public override string ToString()
        {
            return _utf8;
        }
    }
}
