using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace DragonChat
{
    public partial class BZNRichTextBox : RichTextBox
    {
        [DllImport("user32.dll", EntryPoint = "SendMessageA")]
        static extern uint SendMessage(HandleRef hWnd, uint wMsg, uint wParam, uint lParam);

        [DllImport("user32.dll")]
        protected static extern bool HideCaret(IntPtr hWnd);

        const uint WM_USER = 0x400;
        const uint EM_SETEVENTMASK = (WM_USER + 69);
        const uint EM_SETLANGOPTIONS = (WM_USER + 120);
        const uint EM_GETLANGOPTIONS = (WM_USER + 121);

        const int SB_BOTTOM = 7;
        const int WM_VSCROLL = 0x115; 

        const uint WM_SETREDRAW = 0x0B;

        const uint IMF_AUTOFONT = 0x0002;

        int updating = 0;
        uint oldEventMask = 0;

        const int WM_MOUSEMOVE = 0x0200;
        const int WM_LBUTTONDBLCLK = 0x0203;
        const int MK_LBUTTON = 0x0001;
        const int MK_RBUTTON = 0x0002;
        bool selecting = false;
        int startingSelectionStart = -1;

        public void ScrollToBottom()
        {
            SendMessage(new HandleRef(this, Handle), WM_VSCROLL, SB_BOTTOM, 0);
        }

        public void BeginUpdate()
        {
            ++updating;

            if (updating > 1)
                return;

            oldEventMask = SendMessage(new HandleRef(this, Handle), EM_SETEVENTMASK, 0, 0);

            SendMessage(new HandleRef(this, Handle), WM_SETREDRAW, 0, 0);
        }

        public void EndUpdate()
        {
            --updating;

            if (updating > 0)
                return;

            SendMessage(new HandleRef(this, Handle), WM_SETREDRAW, 1, 0);

            SendMessage(new HandleRef(this, Handle), EM_SETEVENTMASK, 0, oldEventMask);
        }

        public void AppendStyle(string text)
        {
            // Regex object for catching style tags.
            Regex r = new Regex(@"\[(?:/)?style([^\]]+)?\]", RegexOptions.IgnoreCase);
            // Int value representing position in the string.
            int pos = 0;
            // Save current values for cursor position and any selected text.
            int SelStart = SelectionStart;
            int SelLen = SelectionLength;
            // Create font object and set initial value to the Font.
            Font font = Font;
            // Create stack object for font objects and add the initial font.
            Stack Fonts = new Stack();
            Fonts.Push(font);
            // Create color object and set initial value to the ForeColor.
            Color color = ForeColor;
            // Create stack collection for color objects and add the initial color.
            Stack Colors = new Stack();
            Colors.Push(color);
            // Create objects for the font value types: FontFamily, Italic, and Bold
            // Set initial values to that of the SelectionFont properties
            FontFamily ff = Font.FontFamily;
            bool italic = Font.Italic;
            bool bold = Font.Bold;
            // Halt drawing until text processing is finished.
            BeginUpdate();
            // Set Indention from edge of control
            this.SelectionIndent = 5;
            // Execute regex pattern matching.
            foreach (Match m in r.Matches(text))
            {
                // If the start of the current match is not equal to the current position in the string.
                if (m.Index != pos)
                {
                    // Append preceeding text or text that appears after a closing style tag.
                    AppendText(text.Substring(pos, m.Index - pos));
                    SelectionStart = TextLength - text.Substring(pos, m.Index - pos).Length;
                    SelectionLength = text.Substring(pos, m.Index - pos).Length;
                    SelectionFont = (Font)Fonts.Peek();
                    SelectionColor = (Color)Colors.Peek();
                }
                // Set current position at the end of the current match.
                pos = m.Index + m.Length;
                // If current match is a closing style tag.
                if (m.Groups[0].Value.Substring(0, 2) == "[/")
                {
                    // If there is more than 1 font left in the collection
                    if (Fonts.Count > 1)
                    {
                        // Remove the top-most font item in the stacks. Set values to the Selection. 
                        SelectionFont = (Font)Fonts.Pop();
                        SelectionColor = (Color)Colors.Pop();
                    }
                }
                // If current match is an opening tag.
                else
                {
                    // Set font and color objects to the current Selection.
                    font = SelectionFont;
                    color = SelectionColor;
                    //bold = SelectionFont.Bold;
                    //italic = SelectionFont.Italic;
                    bold = false;
                    italic = false;
                    // Create array of the properties of the style tag. Parse with semi-colon which represents end of a property.
                    string[] fp = m.Groups[1].Value.Split(new string[] { ";" }, StringSplitOptions.RemoveEmptyEntries);
                    // Loop through properties checking values.
                    for (int i = 0; i < fp.Length; i++)
                    {
                        fp[i] = fp[i].ToLower().Trim();
                        if (fp[i] == "b")
                            bold = true;
                        else if (fp[i] == "i")
                            italic = true;
                        else
                            if (fp[i].Length > 3)
                            {
                                if (fp[i].Substring(0, 3) == "co:")
                                {
                                    try
                                    {
                                        // Convert hexedecimal color value to .Net Color object
                                        color = ColorTranslator.FromHtml(fp[i].Substring(3).Trim());
                                    }
                                    catch (Exception)
                                    {
                                        // If invalid value or error set the last known color.
                                        color = SelectionColor;
                                    }
                                }
                                if (fp[i].Substring(0, 3) == "ff:")
                                {
                                    try
                                    {
                                        // Create fontfamily object using the name of font.
                                        ff = new FontFamily(fp[i].Substring(3).Trim());
                                    }
                                    catch (Exception)
                                    {
                                        // If font does not exist or error set to the last known fontfamily.
                                        ff = SelectionFont.FontFamily;
                                    }
                                }
                            }
                    }
                    // Create new font object and add to font collection.
                    Fonts.Push(MakeFont(ff, bold, italic));
                    // Add edited color object to color collection.
                    Colors.Push(color);
                }
            }
            // If any text exists outside the last pattern match.
            if (pos != text.Length)
            {
                // Append text and apply last known font and color.
                AppendText(text.Substring(pos));
                SelectionStart = TextLength - text.Substring(pos).Length;
                SelectionLength = text.Substring(pos).Length;
                SelectionFont = (Font)Fonts.Peek();
                SelectionColor = (Color)Colors.Peek();
            }
            // Reset cursor position.
            SelectionStart = SelStart;
            SelectionLength = SelLen;
            // Resume drawing.
            EndUpdate();
        }

        private Font MakeFont(FontFamily ff, bool bold, bool italic)
        {
            FontStyle fs = FontStyle.Regular;

            int c = 0, b = 0, i = 0;

            if (ff.IsStyleAvailable(FontStyle.Regular))
                c += 1;
            if (ff.IsStyleAvailable(FontStyle.Bold))
            {
                c += 1;
                b += 1;
            }
            if (ff.IsStyleAvailable(FontStyle.Italic))
            {
                c += 1;
                i += 1;
            }
            if (ff.IsStyleAvailable(FontStyle.Bold | FontStyle.Italic))
            {
                c += 1;
                b += 1;
                i += 1;
            }

            if (c == b)
            {
                fs = fs | FontStyle.Bold;
                bold = false;
            }
            if (c == i)
            {
                fs = fs | FontStyle.Italic;
                italic = false;
            }

            if (bold)
            {
                if (ff.IsStyleAvailable(FontStyle.Bold | fs))
                    fs = FontStyle.Bold | fs;
            }

            if (italic)
            {
                if (ff.IsStyleAvailable(FontStyle.Italic | fs))
                    fs = FontStyle.Italic | fs;
            }

            return new Font(ff, Font.Size, fs);
            
        }

        public bool IsMousePointOnSelection(int X, int Y)
        {
            int charIndex = this.GetExactCharIndexFromPosition(new Point(X, Y));
            return (charIndex >= this.SelectionStart && charIndex < this.SelectionStart + this.SelectionLength);
        }

        public int GetExactCharIndexFromPosition(Point pt)
        {
            int charIndex = this.GetCharIndexFromPosition(pt);
            Point charPoint = this.GetPositionFromCharIndex(charIndex);
            if (pt.X < charPoint.X && charIndex > 0)
            {
                charIndex--;
            }
            return charIndex;
        }

        protected override void OnMouseDown(MouseEventArgs e)
        {
            // Process MouseDown event as normal
            base.OnMouseDown(e);
            // If new selection, initialize 
            if (e.Button == MouseButtons.Left && this.SelectionLength == 0)
            {
                selecting = true;
                startingSelectionStart = this.SelectionStart;
            }
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            // If not selecting, do standard processing
            if (!selecting)
            {
                // If Auto Drag Drop feature enabled, show Arrow cursor when on selection
                if (this.EnableAutoDragDrop)
                {
                    this.Cursor = (e.Button == MouseButtons.None && IsMousePointOnSelection(e.X, e.Y))
                    ? Cursors.Arrow : Cursors.IBeam;
                }
                // Do standard processing
                base.OnMouseMove(e);
                return;
            }
            // Get index of nearest character to mouse cursor
            int charIndex = this.GetCharIndexFromPosition(new Point(e.X, e.Y));
            // If last character, increment index where nearer end of line
            if (charIndex == this.Text.Length - 1)
            {
                Point charPoint = this.GetPositionFromCharIndex(charIndex);
                Point newCharPoint = this.GetPositionFromCharIndex(charIndex + 1);
                if (e.X >= (charPoint.X + ((newCharPoint.X - charPoint.X) / 2)))
                {
                    charIndex++;
                }
            }
            // Calculate selection length based on current / starting character positions
            int length = Math.Abs(charIndex - startingSelectionStart);
            // Update selection
            if (charIndex < startingSelectionStart)
            {
                this.Select(charIndex, length);
            }
            else if (this.SelectionStart < startingSelectionStart)
            {
                this.Select(startingSelectionStart, length);
            }
            else
            {
                this.SelectionLength = length;
            }
        }

        protected override void OnMouseUp(MouseEventArgs mevent)
        {
            selecting = false;
            base.OnMouseUp(mevent);
        }

        protected override void WndProc(ref Message m)
        {
            if (m.Msg == WM_MOUSEMOVE)
            {
                // If selecting, handle mouse movement (bypassing common control)
                if (selecting)
                {
                    this.OnMouseMove(CreateMouseEventArgs(ref m, 1));
                    return;
                }
            }
            else if (m.Msg == WM_LBUTTONDBLCLK)
            {
                // Bypass common control
                base.OnMouseDoubleClick(CreateMouseEventArgs(ref m, 2));
                return;
            }

            base.WndProc(ref m);
        }

        MouseEventArgs CreateMouseEventArgs(ref Message m, int clicks)
        {
            // Get button(s) pressed
            MouseButtons buttons = MouseButtons.None;

            if (((int)m.WParam & MK_LBUTTON) != 0)
            {
                buttons |= MouseButtons.Left;
            }

            if (((int)m.WParam & MK_RBUTTON) != 0)
            {
                buttons |= MouseButtons.Right;
            }

            // Get client co-ordinates
            int x = (int)m.LParam & 0xffff;
            int y = (int)m.LParam >> 16;

            // Return object
            return new MouseEventArgs(buttons, clicks, x, y, 0);
        }

        public BZNRichTextBox()
        {
            InitializeComponent();
            uint lres = SendMessage(new HandleRef(this, Handle), EM_GETLANGOPTIONS, 0, 0);
            lres = lres & ~IMF_AUTOFONT;
            SendMessage(new HandleRef(this, Handle), EM_SETLANGOPTIONS, 0, lres);

            HideCaret(this.Handle);
        }
    }
}
