﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

namespace DragonChat
{
    /// <summary>
    /// Interaction logic for BuzzenDisplay.xaml
    /// </summary>
    public partial class BuzzenDisplay : RichTextBox
    {
        private Dictionary<string, Emoticon> mappings;

        public BuzzenDisplay()
        {
            InitializeComponent();
            this.VerticalScrollBarVisibility = ScrollBarVisibility.Auto;
            this.AutoWordSelection = false;
            this.IsReadOnly = true;
            this.Document.Blocks.Clear();
            this.FontFamily = new FontFamily("Tahoma");
            this.FontSize = 12.5;
            this.Cursor = Cursors.Arrow;
            mappings = new Dictionary<string, Emoticon>();
            mappings.Add(":)", new Emoticon("1.bmp", 26, 24));
            mappings.Add(":d", new Emoticon("2.bmp", 25, 24));
            mappings.Add("*-)", new Emoticon("3.bmp", 25, 24));
            mappings.Add("::(", new Emoticon("13.bmp", 25, 24));
            mappings.Add(":(", new Emoticon("4.bmp", 25, 24));
            mappings.Add(":'(", new Emoticon("5.bmp", 29, 25));
            mappings.Add("(us)", new Emoticon("6.bmp", 24, 24));
            mappings.Add("(a)", new Emoticon("7.bmp", 25, 25));
            mappings.Add("^o)", new Emoticon("8.bmp", 26, 24));
            mappings.Add(":|", new Emoticon("9.bmp", 24, 24));
            mappings.Add(":s", new Emoticon("10.bmp", 25, 24));
            mappings.Add(":o", new Emoticon("11.bmp", 26, 24));
            mappings.Add(":*", new Emoticon("12.bmp", 25, 24));
            mappings.Add(":^)", new Emoticon("14.bmp", 24, 24));
            mappings.Add(":p", new Emoticon("15.bmp", 24, 26));
            mappings.Add("~o", new Emoticon("16.bmp", 26, 25));
            mappings.Add(";)", new Emoticon("17.bmp", 24, 24));
            mappings.Add("(h)", new Emoticon("18.bmp", 25, 24));
            mappings.Add("8o|", new Emoticon("19.bmp", 24, 24));
            mappings.Add("ys)", new Emoticon("20.bmp", 25, 24));
            mappings.Add("|n)", new Emoticon("21.bmp", 25, 24));
            mappings.Add("|^:", new Emoticon("22.bmp", 24, 23));
            mappings.Add("8-)", new Emoticon("23.bmp", 24, 25));
            mappings.Add("8-|", new Emoticon("24.bmp", 25, 23));
            mappings.Add(":@", new Emoticon("25.bmp", 24, 23));
            mappings.Add("(6)", new Emoticon("26.bmp", 28, 25));
            mappings.Add("|-)", new Emoticon("27.bmp", 28, 24));
            mappings.Add(":$", new Emoticon("28.bmp", 23, 25));
            mappings.Add(":-#", new Emoticon("29.bmp", 24, 24));
            mappings.Add("(~)", new Emoticon("30.bmp", 25, 24));
            mappings.Add("+o(", new Emoticon("31.bmp", 24, 23));
            mappings.Add(":~", new Emoticon("32.bmp", 41, 24));
            mappings.Add(":g)", new Emoticon("33.bmp", 28, 27));
            mappings.Add(":h)", new Emoticon("34.bmp", 26, 23));
            mappings.Add("(fl)", new Emoticon("35.bmp", 34, 24));
            mappings.Add("(i)", new Emoticon("36.bmp", 22, 24));
            mappings.Add("(hf)", new Emoticon("37.bmp", 40, 24));
            mappings.Add(":rb", new Emoticon("38.bmp", 25, 24));
            mappings.Add(":w:", new Emoticon("39.bmp", 30, 24));
            mappings.Add("#sp", new Emoticon("40.bmp", 23, 23));
            mappings.Add("^^:", new Emoticon("41.bmp", 52, 30));
            mappings.Add("|~", new Emoticon("42.bmp", 25, 37));
            mappings.Add("(bs)", new Emoticon("43.bmp", 26, 25));
            mappings.Add("(mo)", new Emoticon("44.bmp", 25, 25));
            mappings.Add("(gh)", new Emoticon("45.bmp", 26, 25));
            mappings.Add("(dk)", new Emoticon("46.bmp", 38, 24));
            mappings.Add("(al)", new Emoticon("47.bmp", 26, 24));
            mappings.Add("(g)", new Emoticon("48.bmp", 36, 25));
            mappings.Add(":br", new Emoticon("49.bmp", 27, 34));
            mappings.Add("(brb)", new Emoticon("50.bmp", 53, 25));
            mappings.Add("(as)", new Emoticon("51.bmp", 51, 28));
            mappings.Add("({)", new Emoticon("52.bmp", 24, 23));
            mappings.Add("(})", new Emoticon("53.bmp", 25, 24));
            mappings.Add("(ks)", new Emoticon("54.bmp", 44, 18));
            mappings.Add("(k)", new Emoticon("55.bmp", 25, 17));
            mappings.Add("(l)", new Emoticon("56.bmp", 26, 22));
            mappings.Add("(u)", new Emoticon("57.bmp", 26, 22));
            mappings.Add("(au)", new Emoticon("58.bmp", 43, 23));
            mappings.Add("(@)", new Emoticon("59.bmp", 40, 22));
            mappings.Add("(&)", new Emoticon("60.bmp", 28, 22));
            mappings.Add("(tu)", new Emoticon("61.bmp", 40, 21));
            mappings.Add("(s)", new Emoticon("62.bmp", 24, 23));
            mappings.Add("(r*)", new Emoticon("63.bmp", 36, 22));
            mappings.Add("(*)", new Emoticon("64.bmp", 38, 21));
            mappings.Add("~#", new Emoticon("65.bmp", 38, 28));
            mappings.Add("|w", new Emoticon("66.bmp", 37, 27));
            mappings.Add("(x)", new Emoticon("67.bmp", 23, 23));
            mappings.Add("(z)", new Emoticon("68.bmp", 22, 23));
            mappings.Add("(ff)", new Emoticon("69.bmp", 27, 28));
            mappings.Add("(f)", new Emoticon("70.bmp", 21, 26));
            mappings.Add("(w)", new Emoticon("71.bmp", 26, 22));
            mappings.Add("(d)", new Emoticon("72.bmp", 31, 31));
            mappings.Add("(b)", new Emoticon("73.bmp", 24, 25));
            mappings.Add("(c)", new Emoticon("74.bmp", 30, 24));
            mappings.Add("(pi)", new Emoticon("75.bmp", 34, 21));
            mappings.Add("(%)", new Emoticon("76.bmp", 26, 22));
            mappings.Add("(mp)", new Emoticon("77.bmp", 32, 26));
            mappings.Add("(so)", new Emoticon("78.bmp", 23, 23));
            mappings.Add("(8)", new Emoticon("79.bmp", 23, 22));
            mappings.Add("(e)", new Emoticon("80.bmp", 30, 25));
            mappings.Add("(p)", new Emoticon("81.bmp", 30, 21));
            mappings.Add("(ci)", new Emoticon("82.bmp", 32, 20));
            mappings.Add("(m)", new Emoticon("83.bmp", 23, 26));
            mappings.Add("(?)", new Emoticon("84.bmp", 50, 26));
            mappings.Add("(pl)", new Emoticon("85.bmp", 45, 23));
            mappings.Add("(yn)", new Emoticon("86.bmp", 23, 26));
            mappings.Add("(y)", new Emoticon("87.bmp", 24, 23));
            mappings.Add("(n)", new Emoticon("88.bmp", 25, 22));
            mappings.Add("(o)", new Emoticon("89.bmp", 23, 23));
            mappings.Add("(st)", new Emoticon("90.bmp", 31, 27));
            mappings.Add("(rn)", new Emoticon("91.bmp", 30, 25));
            mappings.Add("(**)", new Emoticon("92.bmp", 30, 25));
            mappings.Add("(um)", new Emoticon("93.bmp", 27, 24));
            mappings.Add("(#)", new Emoticon("94.bmp", 22, 22));
            mappings.Add("(ip)", new Emoticon("95.bmp", 35, 24));
            mappings.Add("(t)", new Emoticon("96.bmp", 28, 22));
            mappings.Add(":[", new Emoticon("97.bmp", 41, 22));
            mappings.Add("(ap)", new Emoticon("98.bmp", 43, 24));
            mappings.Add("-sc", new Emoticon("99.bmp", 33, 28));
            mappings.Add("(r)", new Emoticon("100.bmp", 27, 22));
            mappings.Add("(bu)", new Emoticon("101.bmp", 26, 26));
            mappings.Add("(^^)", new Emoticon("102.bmp", 31, 28));
            mappings.Add("(^)", new Emoticon("103.bmp", 30, 28));
            mappings.Add("-@)", new Emoticon("104.bmp", 48, 29));
            mappings.Add(":y", new Emoticon("105.bmp", 25, 23));
            mappings.Add("(wt)", new Emoticon("106.bmp", 33, 32));
            mappings.Add("(tk)", new Emoticon("107.bmp", 27, 30));
            mappings.Add("(ed)", new Emoticon("108.bmp", 32, 31));
            mappings.Add("(b^)", new Emoticon("109.bmp", 33, 29));
            mappings.Add(":_)", new Emoticon("110.bmp", 25, 25));
            mappings.Add("(sm)", new Emoticon("111.bmp", 29, 30));
        }

        public void Append(string text)
        {
            TextRange tr = new TextRange(this.Document.ContentStart, this.Document.ContentEnd);

            // Regex object for catching style tags.
            Regex r = new Regex(@"\[(?:/)?style([^\]]+)?\]", RegexOptions.IgnoreCase);
            // Int value representing position in the string.
            int pos = 0;
            // Inserted Text Length
            int iLen = 0;
            // Create style stack and insert default
            Stack Styles = new Stack();
            StyleTag Style = new StyleTag(this.FontFamily, Brushes.Black, false, false);
            Styles.Push(Style);
            BrushConverter bc = new BrushConverter();
            foreach (Match m in r.Matches(text))
            {
                tr = new TextRange(this.Document.ContentEnd, this.Document.ContentEnd);
                // If the start of the current match is not equal to the current position in the string.
                if (m.Index != pos)
                {
                    // Append preceeding text or text that appears after a closing style tag.
                    tr.Text = text.Substring(pos, m.Index - pos);
                    iLen += tr.Text.Length;
                    Style = (StyleTag)Styles.Peek();
                    tr.ApplyPropertyValue(TextElement.FontFamilyProperty, Style.FontFamily);
                    tr.ApplyPropertyValue(TextElement.ForegroundProperty, Style.ForeColor);
                    tr.ApplyPropertyValue(TextElement.FontStyleProperty, Style.IsItalic ? FontStyles.Italic : FontStyles.Normal);
                    tr.ApplyPropertyValue(TextElement.FontWeightProperty, Style.IsBold ? FontWeights.Bold : FontWeights.Normal);
                    tr.ApplyPropertyValue(TextElement.FontSizeProperty, this.FontSize);
                }
                // Set current position at the end of the current match.
                pos = m.Index + m.Length;
                // If current match is a closing style tag.
                if (m.Groups[0].Value.Substring(0, 2) == "[/")
                {   // If there is more than 1 font left in the collection
                    if (Styles.Count > 1)
                        Styles.Pop();
                }
                // If current match is an opening tag.
                else
                {
                    bool bold = false;
                    bool italic = false;

                    Style = (StyleTag)Styles.Peek();

                    System.Windows.Media.SolidColorBrush _forecolor = Style.ForeColor;
                    System.Windows.Media.FontFamily _fontfamily = Style.FontFamily;

                    // Create array of the properties of the style tag. Parse with semi-colon which represents end of a property.
                    string[] fp = m.Groups[1].Value.Split(new string[] { ";" }, StringSplitOptions.RemoveEmptyEntries);
                    // Loop through properties checking values.
                    for (int i = 0; i < fp.Length; i++)
                    {
                        fp[i] = fp[i].ToLower().Trim();
                        if (fp[i] == "b")
                            bold = true;
                        else if (fp[i] == "i")
                            italic = true;
                        else
                            if (fp[i].Length > 3)
                            {
                                if (fp[i].Substring(0, 3) == "co:")
                                {
                                    try
                                    {
                                        // Convert hexedecimal color value to .Net Color object
                                        _forecolor = (SolidColorBrush)bc.ConvertFrom(fp[i].Substring(3).Trim());
                                    }
                                    catch (Exception)
                                    {
                                        // If invalid value leave as last known value
                                    }
                                }
                                if (fp[i].Substring(0, 3) == "ff:")
                                {
                                    try
                                    {
                                        // Create fontfamily object using the name of font.
                                        _fontfamily = new FontFamily(fp[i].Substring(3).Trim());
                                    }
                                    catch (Exception)
                                    {
                                        // If font does not exist or error set to the last known fontfamily.
                                    }
                                }
                            }
                    }
                    // Create new style object and add to style collection.
                    Styles.Push(new StyleTag(_fontfamily, _forecolor, bold, italic));
                }
            }
            // If any text exists outside the last pattern match.
            if (pos != text.Length)
            {
                // Append text and apply last known font and color.
                tr = new TextRange(this.Document.ContentEnd, this.Document.ContentEnd);
                tr.Text = text.Substring(pos);
                iLen += tr.Text.Length;
                Style = (StyleTag)Styles.Peek();
                tr.ApplyPropertyValue(TextElement.FontFamilyProperty, Style.FontFamily);
                tr.ApplyPropertyValue(TextElement.ForegroundProperty, Style.ForeColor);
                tr.ApplyPropertyValue(TextElement.FontStyleProperty, Style.IsItalic ? FontStyles.Italic : FontStyles.Normal);
                tr.ApplyPropertyValue(TextElement.FontWeightProperty, Style.IsBold ? FontWeights.Bold : FontWeights.Normal);
                tr.ApplyPropertyValue(TextElement.FontSizeProperty, this.FontSize);
            }

            tr = new TextRange(this.Document.ContentStart, this.Document.ContentEnd);
            TextPointer stp = this.Document.ContentStart.GetPositionAtOffset(tr.Text.Length - iLen);

            try
            {
                InsertHyperlink(stp);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "InsertHyperlink error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
            try
            {
                InsertEmoticons(stp);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "InsertEmoticons error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        void h_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            if (e.ClickCount == 2)
                System.Diagnostics.Process.Start(((Hyperlink)sender).Tag.ToString());
        }

        private void InsertEmoticons(TextPointer position)
        {
            string match = string.Empty;
            TextRange tr;
            BitmapImage bi;
            while (position != null)
            {
                if (position.CompareTo(this.Document.ContentEnd) == 0)
                {
                    break;
                }
                if (position.GetPointerContext(LogicalDirection.Forward) == TextPointerContext.Text)
                {
                    String text = position.GetTextInRun(LogicalDirection.Forward);
                    Int32 indexInRun = -1;
                    match = string.Empty;
                    foreach (KeyValuePair<string, Emoticon> pair in mappings)
                    {
                        if (text.ToLower().Contains(pair.Key))
                        {
                            int newPosition = text.ToLower().IndexOf(pair.Key);
                            if (indexInRun == -1)
                            {
                                match = pair.Key;
                                indexInRun = newPosition;
                            }
                            if (newPosition < indexInRun)
                            {
                                match = pair.Key;
                                indexInRun = newPosition;
                            }
                        }
                    }

                    if (indexInRun >= 0)
                    {
                        position = position.GetPositionAtOffset(indexInRun);
                        tr = new TextRange(position, position.GetPositionAtOffset(match.Length));

                        tr.Text = string.Empty;
                        Emoticon emot = (Emoticon)mappings[match];
                        bi = new BitmapImage();
                        bi.BeginInit();
                        bi.UriSource = new Uri("pack://application:,,,/Resources/" + emot.Image);
                        bi.DecodePixelWidth = emot.Width;
                        bi.DecodePixelHeight = emot.Height;
                        bi.EndInit();

                        System.Windows.Controls.Image image = new System.Windows.Controls.Image();
                        image.Width = emot.Width;
                        image.Height = emot.Height;
                        image.Stretch = System.Windows.Media.Stretch.None;
                        image.Source = bi;

                        new InlineUIContainer(image, position);
                    }
                    else
                    {
                        // If a match is not found, go over to the next context position after the "textRun".
                        position = position.GetPositionAtOffset(text.Length);
                    }
                }
                else
                {
                    //If the current position doesn't represent a text context position, go to the next context position.
                    // This can effectively ignore the formatting or embed element symbols.
                    position = position.GetNextContextPosition(LogicalDirection.Forward);
                }
            }
        }

        private void InsertHyperlink(TextPointer position)
        {
            string match = string.Empty;
            Regex r = new Regex("(?:^|[\\s\\[\\]\\}\\{\\(\\)\\\'\\\"<>])((?:(?:https?|gopher|ftp|file|irc):\\/\\/|www\\.)[a-zA-Z0-9\\.\\-=;&%\\?]+(?:\\/?[a-zA-Z0-9\\.\\-=;&%\\?#]*)*)");

            Hyperlink h;

            while (position != null)
            {
                if (position.CompareTo(this.Document.ContentEnd) == 0)
                {
                    break;
                }
                if (position.GetPointerContext(LogicalDirection.Forward) == TextPointerContext.Text)
                {
                    String text = position.GetTextInRun(LogicalDirection.Forward);
                    Int32 indexInRun = -1;
                    if (r.IsMatch(text))
                    {
                        Match m = r.Match(text);
                        match = m.Groups[1].Value;
                        indexInRun = m.Groups[1].Index;
                    }

                    if (indexInRun >= 0)
                    {
                        position = position.GetPositionAtOffset(indexInRun);
                        h = new Hyperlink(position, position.GetPositionAtOffset(match.Length));
                        h.Tag = match;
                        h.Foreground = Brushes.Blue;
                        h.TextDecorations = null;
                        h.Cursor = System.Windows.Input.Cursors.Hand;
                        h.MouseLeftButtonDown += new MouseButtonEventHandler(h_MouseLeftButtonDown);
                        position = position.GetPositionAtOffset(match.Length);
                    }
                    else
                    {
                        position = position.GetPositionAtOffset(text.Length);
                    }
                }
                else
                {
                    position = position.GetNextContextPosition(LogicalDirection.Forward);
                }
            }
        }
    }


    public struct StyleTag
    {
        bool _bold;
        bool _italic;

        System.Windows.Media.SolidColorBrush _forecolor;
        System.Windows.Media.FontFamily _fontfamily;

        public StyleTag(System.Windows.Media.FontFamily fontfamily, System.Windows.Media.SolidColorBrush forecolor, bool bold, bool italic)
        {
            this._fontfamily = fontfamily;
            this._forecolor = forecolor;
            this._bold = bold;
            this._italic = italic;
        }

        public bool IsItalic
        {
            get
            {
                return this._italic;
            }
        }

        public bool IsBold
        {
            get
            {
                return this._bold;
            }
        }

        public System.Windows.Media.SolidColorBrush ForeColor
        {
            get
            {
                return this._forecolor;
            }
        }

        public System.Windows.Media.FontFamily FontFamily
        {
            get
            {
                return this._fontfamily;
            }
        }
    }

    public struct Emoticon
    {
        string _image;
        int _height;
        int _width;

        public Emoticon(string image, int width, int height)
        {
            _image = image;
            _height = height;
            _width = width;
        }

        public string Image
        {
            get
            {
                return _image;
            }
        }

        public int Height
        {
            get
            {
                return _height;
            }
        }

        public int Width
        {
            get
            {
                return _width;
            }
        }
    }
}
