﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace DragonChat
{
    public partial class AccessList : Form
    {
        Client client;
        AccessCollection alist;
        AddAccess AddEntryDialog;

        delegate void CmdCallback();

        public AccessList(Client p)
        {
            InitializeComponent();
            client = p;
            this.Load += new EventHandler(AccessList_Load);
            this.lvAccess.ListViewItemSorter = new ALSort();
            this.lvAccess.SelectedIndexChanged += new EventHandler(lvAccess_SelectedIndexChanged);
        }
        public AccessCollection List
        {
            get
            {
                return alist;
            }
            set
            {
                this.alist = value;
            }
        }

        void AccessList_Load(object sender, EventArgs e)
        {
            //UpdateList();
        }

        void lvAccess_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (lvAccess.SelectedItems.Count > 0)
            {
                rtbTag.Text = ((AccessEntry)lvAccess.SelectedItems[0].Tag).Tag;
            }
            else
            {
                rtbTag.Clear();
            }
        }

        public void UpdateList()
        {
            if (lvAccess.InvokeRequired)
            {
                CmdCallback d = new CmdCallback(this.UpdateList);
                this.Invoke(d);
            }
            else
            {
                try
                {
                    lvAccess.BeginUpdate();
                    lvAccess.Items.Clear();
                    rtbTag.Clear();
                    this.Text = "Access List";
                    AccessEntry[] Entries = alist.GetList();
                    if (Entries != null)
                    {
                        ListViewItem item;
                        ListViewItem.ListViewSubItem Setby;
                        ListViewItem.ListViewSubItem Level;
                        ListViewItem.ListViewSubItem Duration;
                        StringBuilder sb = new StringBuilder();
                        for (int i = 0; i < Entries.Length; i++)
                        {
                            sb.Append(Entries[i].Entry);
                            item = new ListViewItem(Entries[i].Entry);
                            item.ToolTipText = sb.ToString();
                            item.Tag = Entries[i];
                            Setby = new ListViewItem.ListViewSubItem(item, Entries[i].SetbyNick);
                            Level = new ListViewItem.ListViewSubItem(item, Entries[i].Level);
                            Duration = new ListViewItem.ListViewSubItem(item, Entries[i].Duration);
                            item.SubItems.Add(Setby);
                            item.SubItems.Add(Level);
                            item.SubItems.Add(Duration);
                            switch (Entries[i].Level)
                            {
                                case "OWNER":
                                    if (!ownerToolStripMenuItem.Checked)
                                        lvAccess.Items.Add(item);
                                    break;
                                case "HOST":
                                    if (!hostStripMenuItem.Checked)
                                        lvAccess.Items.Add(item);
                                    break;
                                case "VOICE":
                                    if (!voiceToolStripMenuItem.Checked)
                                        lvAccess.Items.Add(item);
                                    break;
                                case "GRANT":
                                    if (!grantToolStripMenuItem.Checked)
                                        lvAccess.Items.Add(item);
                                    break;
                                case "DENY":
                                    if (!denyBanToolStripMenuItem.Checked)
                                        lvAccess.Items.Add(item);
                                    break;
                                default:
                                    break;
                            }
                            sb.Remove(0, sb.Length);
                        }

                    }
                    if (lvAccess.Items.Count > 12)
                        hdrEntry.Width = 182;
                    else
                        hdrEntry.Width = 200;
                    this.Text = "Access List - " + lvAccess.Items.Count.ToString() + " entries found for " + Client.Channel;
                    lvAccess.EndUpdate();
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }
            }
        }

        private void addEntryToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (AddEntryDialog == null)
                AddEntryDialog = new AddAccess();
           if (AddEntryDialog.ShowDialog() == DialogResult.OK)
           {
                client.SendData("ACCESS " + Client.Channel + " ADD " + AddEntryDialog.Level + " " + AddEntryDialog.Entry + " " + AddEntryDialog.CommentTag);
           }
        }

        private void deleteEntryToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (lvAccess.SelectedItems.Count == 1)
            {
                client.SendData("ACCESS " + Client.Channel + " DELETE " + lvAccess.SelectedItems[0].SubItems[2].Text + " " + lvAccess.SelectedItems[0].Text);
            }
        }

        private void updateListToolStripMenuItem_Click(object sender, EventArgs e)
        {
            UpdateList();
        }

        private void ownerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            UpdateList();
        }

        private void toolStripMenuItem4_Click(object sender, EventArgs e)
        {
            UpdateList();
        }

        private void voiceToolStripMenuItem_Click(object sender, EventArgs e)
        {
            UpdateList();
        }

        private void grantToolStripMenuItem_Click(object sender, EventArgs e)
        {
            UpdateList();
        }

        private void denyBanToolStripMenuItem_Click(object sender, EventArgs e)
        {
            UpdateList();
        }

        private void ownerToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            client.SendData("ACCESS " + Client.Channel + " CLEAR OWNER");
        }

        private void hostToolStripMenuItem_Click(object sender, EventArgs e)
        {
            client.SendData("ACCESS " + Client.Channel + " CLEAR HOST");
        }

        private void voiceToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            client.SendData("ACCESS " + Client.Channel + " CLEAR VOICE");
        }

        private void grantToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            client.SendData("ACCESS " + Client.Channel + " CLEAR GRANT");
        }

        private void denyBanToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            client.SendData("ACCESS " + Client.Channel + " CLEAR DENY");
        }

        private void allToolStripMenuItem_Click(object sender, EventArgs e)
        {
            client.SendData("ACCESS " + Client.Channel + " CLEAR");
        }
    }
}
