﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.IO;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace DragonChat
{
    public partial class ChannelManager : Form
    {
        private Client client = null;

        private string Topic = string.Empty;
        private string Onjoin = string.Empty;

        public string Ownerkey = string.Empty;
        public string Hostkey = string.Empty;

        public string Category = string.Empty;
        private int CategoryIndex = 0;

        private bool IsTopic = true;

        private string limit = string.Empty;
        private string rkey = string.Empty;

        public ChannelManager(Client c)
        {
            InitializeComponent();
            client = c;
            Topic = c.Topic;
            Onjoin = c.OnjoinMessage;
            rtbDisplay.Click += new EventHandler(rtbDisplay_Click);
            rtbMessage.MouseEnter += new EventHandler(rtbMessage_MouseEnter);
            rtbMessage.MouseLeave += new EventHandler(rtbMessage_MouseLeave);
            this.MouseMove += new MouseEventHandler(ChannelManager_MouseMove);
            this.MouseDown += new MouseEventHandler(ChannelManager_MouseDown);
        }

        void ChannelManager_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.X < rtbMessage.Left || e.Y < rtbMessage.Top || e.X >= rtbMessage.Width || e.Y >= rtbMessage.Height)
            {
                string msg = rtbMessage.Text.Replace("\r\n", "[br]");
                if (IsTopic)
                    Topic = msg;
                else
                    Onjoin = msg;
                rtbMessage.Visible = false;
                rtbDisplay.Text = "";
                rtbDisplay.AppendStyle(msg.Replace("[br]","\f"));
                rtbDisplay.Visible = true;
            }
        }

        void ChannelManager_MouseMove(object sender, MouseEventArgs e)
        {
            if (e.X < 0 || e.Y < 0 || e.X >= Width || e.Y >= Height)
                this.Capture = false;
            else if (rtbMessage.Visible)
                this.Capture = true;
        }

        void rtbMessage_MouseLeave(object sender, EventArgs e)
        {
            if (rtbMessage.Visible)
                this.Capture = true;
        }

        void rtbMessage_MouseEnter(object sender, EventArgs e)
        {
            this.Capture = false;
        }

        void rtbDisplay_Click(object sender, EventArgs e)
        {
            rtbDisplay.Visible = false;
            rtbMessage.Visible = true;
        }

        private void ChannelManager_Load(object sender, EventArgs e)
        {
            UpdateSettings();
            this.tslTopic_Click(new object(), new EventArgs());
        }

        public void UpdateSettings()
        {
            chkFlood.Checked = false;
            chkGuestWhispers.Checked = false;
            chkInvite.Checked = false;
            chkKey.Checked = false;
            txtKey.Text = string.Empty;
            txtLimit.Text = string.Empty;
            chkModerate.Checked = false;
            chkPrivate.Checked = false;
            chkSecret.Checked = false;
            chkWhispers.Checked = false;
            Topic = client.Topic;
            Onjoin = client.OnjoinMessage;
            string[] m = client.Modes.Split(new string[] { " " }, StringSplitOptions.RemoveEmptyEntries);
            foreach (Char c in m[0].ToCharArray())
            {
                switch (c)
                {
                    case 'i':
                        chkInvite.Checked = true;
                        break;
                    case 'm':
                        chkModerate.Checked = true;
                        break;
                    case 'P':
                        chkFlood.Checked = true;
                        break;
                    case 'p':
                        chkPrivate.Checked = true;
                        break;
                    case 's':
                        chkSecret.Checked = true;
                        break;
                    case 'w':
                        chkWhispers.Checked = true;
                        break;
                    case 'W':
                        chkGuestWhispers.Checked = true;
                        break;
                    case 'k':
                        chkKey.Checked = true;
                        rkey = m[2];
                        txtKey.Text = rkey;
                        break;
                    case 'l':
                        limit = m[1];
                        txtLimit.Text = limit;
                        break;
                }
            }
            rtbOwner.Text = client.Encode(Ownerkey);
            rtbHost.Text = client.Encode(Hostkey);
            CategoryText(Category);
            this.tslTopic_Click(new object(), new EventArgs());
        }

        private void CategoryText(string abbrev)
        {
            if (abbrev == string.Empty)
                cboCategory.SelectedIndex = 0;
            else
            {
                switch (abbrev.ToUpper())
                {
                    case "CC":
                        cboCategory.SelectedIndex = 1;
                        break;
                    case "CP":
                        cboCategory.SelectedIndex = 2;
                        break;
                    case "GN":
                        cboCategory.SelectedIndex = 4;
                        break;
                    case "UL":
                        cboCategory.SelectedIndex = 14;
                        break;
                    case "IN":
                        cboCategory.SelectedIndex = 5;
                        break;
                    case "TN":
                        cboCategory.SelectedIndex = 13;
                        break;
                    case "RM":
                        cboCategory.SelectedIndex = 11;
                        break;
                    case "SP":
                        cboCategory.SelectedIndex = 12;
                        break;
                    case "LS":
                        cboCategory.SelectedIndex = 6;
                        break;
                    case "RP":
                        cboCategory.SelectedIndex = 10;
                        break;
                    case "RL":
                        cboCategory.SelectedIndex = 9;
                        break;
                    case "ET":
                        cboCategory.SelectedIndex = 3;
                        break;
                    case "NE":
                        cboCategory.SelectedIndex = 7;
                        break;
                    case "PR":
                        cboCategory.SelectedIndex = 8;
                        break;
                    default:
                        cboCategory.SelectedIndex = 0;
                        break;
                }
                CategoryIndex = cboCategory.SelectedIndex;
            }
        }

        private void tslTopic_Click(object sender, EventArgs e)
        {
            IsTopic = true;
            rtbMessage.Text = Topic;
            rtbDisplay.Text = "";
            rtbDisplay.AppendStyle(Topic.Replace("[br]","\f"));
            rtbMessage.Visible = false;
            rtbDisplay.Visible = true;
            tslOnjoin.Font = new Font(tslOnjoin.Font, FontStyle.Regular);
            tslTopic.Font = new Font(tslTopic.Font, FontStyle.Bold);
        }

        private void tslOnjoin_Click(object sender, EventArgs e)
        {
            IsTopic = false;
            rtbMessage.Text = Onjoin;
            rtbDisplay.Text = "";
            rtbDisplay.AppendStyle(Onjoin.Replace("[br]", "\f"));
            rtbMessage.Visible = false;
            rtbDisplay.Visible = true;
            tslOnjoin.Font = new Font(tslOnjoin.Font, FontStyle.Bold);
            tslTopic.Font = new Font(tslTopic.Font, FontStyle.Regular);
        }

        private void btnOK_Click(object sender, EventArgs e)
        {
            if (client.IsHost)
            {
                string PROP = "PROP " + Client.Channel;
                if (Topic != client.Topic)
                {
                    client.SendData(PROP + " TOPIC " + Topic);
                }
                if (Onjoin != client.OnjoinMessage)
                {
                    client.SendData(PROP + " ONJOIN " + Onjoin);
                }
                if (CategoryIndex != cboCategory.SelectedIndex)
                {
                    string cat = string.Empty;
                    switch (cboCategory.SelectedIndex)
                    {
                        case 14:
                            cat = "UL";
                            break;
                        case 13:
                            cat = "TN";
                            break;
                        case 12:
                            cat = "SP";
                            break;
                        case 11:
                            cat = "RM";
                            break;
                        case 10:
                            cat = "RP";
                            break;
                        case 9:
                            cat = "RL";
                            break;
                        case 8:
                            cat = "PR";
                            break;
                        case 7:
                            cat = "NE";
                            break;
                        case 6:
                            cat = "LS";
                            break;
                        case 5:
                            cat = "IN";
                            break;
                        case 4:
                            cat = "GN";
                            break;
                        case 3:
                            cat = "ET";
                            break;
                        case 2:
                            cat = "CP";
                            break;
                        case 1:
                            cat = "CC";
                            break;
                    }
                    if (cat != string.Empty)
                    {
                        client.SendData(PROP + " CATEGORY " + cat);
                    }
                }
                if (client.IsOwner)
                {
                    if (client.Encode(Ownerkey) != rtbOwner.Text)
                    {
                        client.SendData(PROP + " OWNERKEY " + rtbOwner.Text);
                    }
                    if (client.Encode(Hostkey) != rtbHost.Text)
                    {
                        client.SendData(PROP + " HOSTKEY " + rtbOwner.Text);
                    }
                }
                string[] m = client.Modes.Split(new string[] { " " }, StringSplitOptions.RemoveEmptyEntries);
                string current_modes = m[0];
                if (m.Length > 1)
                    limit = m[1];
                if (m.Length > 2)
                    rkey = m[2];
                // Channel Key Mode
                if (chkKey.Checked)
                {
                    if (txtKey.Text != string.Empty)
                    {
                        if (current_modes.IndexOf('k') > -1)
                        {
                            if (rkey != txtKey.Text)
                                client.SendData("MODE " + Client.Channel + " +k " + txtKey.Text);
                        }
                        else
                            client.SendData("MODE " + Client.Channel + " +k " + txtKey.Text);
                    }
                }
                else
                {
                    if (current_modes.IndexOf('k') > -1)
                    {
                        client.SendData("MODE " + Client.Channel + " -k " + rkey);
                    }
                }
                // Channel Limit Mode
                if (txtLimit.Text != limit)
                {
                    int l;
                    if (int.TryParse(txtLimit.Text, out l))
                        l = -1;
                    if (-1 < l || l < 101)
                        client.SendData("MODE " + Client.Channel + " +l " + txtLimit.Text);
                }
                // Other Channel Modes
                string am = string.Empty;
                string rm = string.Empty;
                // Invite Only
                if (chkInvite.Checked)
                {
                    if (current_modes.IndexOf('i') == -1)
                        am += "i";
                }
                else
                {
                    if (current_modes.IndexOf('i') > -1)
                        rm += "i";
                }
                // Flood Protection
                if (chkFlood.Checked)
                {
                    if (current_modes.IndexOf('P') == -1)
                        am += "P";
                }
                else
                {
                    if (current_modes.IndexOf('P') > -1)
                        rm += "P";
                }
                // Secret
                if (chkSecret.Checked)
                {
                    if (current_modes.IndexOf('s') == -1)
                        am += "s";
                }
                else
                {
                    if (current_modes.IndexOf('s') > -1)
                        rm += "s";
                }
                // No Whispers
                if (chkWhispers.Checked)
                {
                    if (current_modes.IndexOf('w') == -1)
                        am += "w";
                }
                else
                {
                    if (current_modes.IndexOf('w') > -1)
                        rm += "w";
                }
                // Private
                if (chkPrivate.Checked)
                {
                    if (current_modes.IndexOf('p') == -1)
                        am += "p";
                }
                else
                {
                    if (current_modes.IndexOf('p') > -1)
                        rm += "p";
                }
                // No Guest Whispers
                if (chkGuestWhispers.Checked)
                {
                    if (current_modes.IndexOf('W') == -1)
                        am += "W";
                }
                else
                {
                    if (current_modes.IndexOf('W') > -1)
                        rm += "W";
                }
                // Moderate
                if (chkModerate.Checked)
                {
                    if (current_modes.IndexOf('m') == -1)
                        am += "m";
                }
                else
                {
                    if (current_modes.IndexOf('m') > -1)
                        rm += "m";
                }
                if (am != string.Empty)
                    client.SendData("MODE " + Client.Channel + " +" + am);
                if (rm != string.Empty)
                    client.SendData("MODE " + Client.Channel + " -" + rm);
            }
        }
    }
}
