using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Net.Sockets;
using System.Text.RegularExpressions;

namespace DragonChat
{
    public partial class Client : Form
    {
        private static Setup setup;
        private static AccessList alist;
        private static ChannelManager chanman;

        private const int WS_CLIPCHILDREN = 0x02000000;
        private const int WS_CLIPSIBLINGS = 0x04000000;

        public static string Nickname = string.Empty;
        public static string Email = string.Empty;
        public static string Password = string.Empty;
        public static string Channel = string.Empty;
        //
        private string Ownerkey = string.Empty;
        private string Hostkey = string.Empty;
        private string Category = string.Empty;
        // Font type, style, and color for text output to server.
        public Font InputFont;
        public System.Drawing.Color InputColor;

        public string Topic = String.Empty;
        public string OnjoinMessage = string.Empty;

        public string Modes = string.Empty;

        public static bool ShowTimeStamp = false;

        private bool AttemptReconnect = false;
        private int ReconnectAttempts = 0;
        private Timer ReconnectTimer = new Timer();

        private bool inchan = false; // In channel?
        private bool rejoin = true; // Rejoin channel when kicked?
        private bool moderated = false; // Channel moderated?

        private string packet;
        private bool flushed = true;

        private string link = string.Empty;

        private bool dclick = false;

        private User snick = null;

        System.Windows.Input.MouseButtonEventArgs PreviewMouseDownArgs = null;

        private Connection Buzzen = new Connection();
        private UserCollection Nicklist = new UserCollection(); // Users in channel 
        private WhisperCollection Whispers = new WhisperCollection(); // Opened whispers.
        private AccessCollection AccessEntries = new AccessCollection(); // List of last know entries in the access list.
        // Callbacks for avoiding cross-thread exceptions
        delegate void SetTextCallback(string text);
        delegate void BoolCallBack(bool isconn);
        delegate void SetColorTextCallback(string text, System.Windows.Media.Color color, bool bold, bool italic);
        delegate void OnTextCallback(string nick, string addr, string text);
        delegate void CmdCallback();
        delegate void AddUserCallback(User user);

        delegate System.Windows.Forms.ListViewItem GetItemCallback(string nick);
        delegate void ChangeIconCallback(System.Windows.Forms.ListViewItem item, int image);
        delegate void ChangeNameCallback(System.Windows.Forms.ListViewItem item, string newnick);
        delegate WhisperForm CreateWhisperCallback(string nickname);

        public Client()
        {
            InitializeComponent();
            InputFont = new Font(rtbOutput.FontFamily.Source, rtbInput.Font.Size, FontStyle.Regular);
            InputColor = System.Drawing.Color.Black;
            string ini = Application.StartupPath + @"\DragonChat.ini";
            IniFile iFile = new IniFile(ini);
            if (File.Exists(ini))
            {
                Email = iFile.GetString("Settings", "Email", string.Empty);
                Password = iFile.GetString("Settings", "Password", string.Empty);
                Nickname = iFile.GetString("Settings", "Nickname", string.Empty);
                Channel = iFile.GetString("Settings", "Channel", string.Empty);
                double size = iFile.GetDouble("Settings", "FontSize", 12.5);
                if (size > 12.5)
                    this.rtbOutput.FontSize = size;
                string sFont = iFile.GetString("Settings", "Font", string.Empty);
                if (sFont != string.Empty)
                {
                    System.Drawing.FontFamily sf;
                    try
                    {
                        sf = new System.Drawing.FontFamily(sFont);
                    }
                    catch (Exception)
                    {
                        sf = new System.Drawing.FontFamily(rtbOutput.FontFamily.Source);
                    }
                    string bold = iFile.GetString("Settings", "Bold", string.Empty);
                    string italic = iFile.GetString("Settings", "Italic", string.Empty);
                    InputFont = MakeFont(sf, bold == "True" ? true : false, italic == "True" ? true : false);
                }
                string sColor = iFile.GetString("Settings", "Color", string.Empty);
                if (sColor != string.Empty)
                {
                    Regex r = new Regex(@"^([\d]?[\d]?[\d]),([\d]?[\d]?[\d]),([\d]?[\d]?[\d])$");
                    Match rm;
                    rm = r.Match(sColor);
                    if (rm.Success)
                    {
                        int ri, gi, bi;
                        if (!int.TryParse(rm.Groups[1].Value, out ri))
                            ri = -1;
                        if (!int.TryParse(rm.Groups[2].Value, out gi))
                            gi = -1;
                        if (!int.TryParse(rm.Groups[3].Value, out bi))
                            bi = -1;
                        if ((ri < 0) || (gi < 0) || (bi < 0))
                        {
                            InputColor = System.Drawing.Color.Black;
                        }
                        else if ((ri > 255) || (gi > 255) || (bi > 255))
                        {
                            InputColor = System.Drawing.Color.Black;
                        }
                        else
                        {
                            InputColor = System.Drawing.Color.FromArgb(ri, gi, bi);
                        }
                    }
                    else
                        InputColor = System.Drawing.Color.Black;
                }
            }
            this.Resize += new EventHandler(Client_Resize);
            this.TrayIcon.DoubleClick += new EventHandler(TrayIcon_DoubleClick);
            // Custom sorting class for nicklist
            this.lvwNicknames.ListViewItemSorter = new NLSort();
            // Addtional control events
            this.rtbInput.KeyDown += new KeyEventHandler(this.rtbInput_KeyDown);
            this.rtbInput.TextChanged += new EventHandler(this.rtbInput_TextChanged);
            // Display
            this.rtbOutput.MouseDoubleClick += new System.Windows.Input.MouseButtonEventHandler(rtbOutput_MouseDoubleClick);
            this.rtbOutput.MouseMove += new System.Windows.Input.MouseEventHandler(rtbOutput_MouseMove);
            this.rtbOutput.PreviewMouseDown += new System.Windows.Input.MouseButtonEventHandler(rtbOutput_PreviewMouseDown);
            this.rtbOutput.AddHandler(System.Windows.UIElement.MouseDownEvent, new System.Windows.RoutedEventHandler(this.rtbOutput_MouseDown), true);
            //
            this.lvwNicknames.MouseClick += new MouseEventHandler(lvwNicknames_MouseClick);
            this.lvwNicknames.MouseDoubleClick += new MouseEventHandler(lvwNicknames_MouseDoubleClick);
            this.lvwNicknames.SelectedIndexChanged += new EventHandler(lvwNicknames_SelectedIndexChanged);
            //
            mnuNicklist.Enabled = false;
            // Socket events
            this.Buzzen.OnConnect += new Connection.ConnectionHandler(this.OnConnect);
            this.Buzzen.OnDisconnect += new Connection.ConnectionHandler(this.OnDisconnect);
            this.Buzzen.OnDataArrival +=new Connection.DataArrivalHandler(this.OnDataArrival);
            // Reconnect Timer
            this.ReconnectTimer.Tick += new EventHandler(ReconnectTimer_Tick);
        }

        void ReconnectTimer_Tick(object sender, EventArgs e)
        {
            SetText("Attempt to reconnect...");
            ReconnectTimer.Stop();
            Connect();
        }

        protected override CreateParams CreateParams
        {
            get
            {
                CreateParams cp = base.CreateParams;
                cp.Style |= WS_CLIPCHILDREN;
                cp.Style |= WS_CLIPSIBLINGS;
                return cp;
            }
        }

        void rtbOutput_MouseDoubleClick(object sender, System.Windows.Input.MouseButtonEventArgs e)
        {
            rtbOutput.Selection.Select(rtbOutput.CaretPosition, rtbOutput.CaretPosition);
            e.Handled = true;
        }

        void rtbOutput_MouseMove(object sender, System.Windows.Input.MouseEventArgs e)
        {
            TextPointer tp = rtbOutput.GetPositionFromPoint(System.Windows.Input.Mouse.GetPosition(this.rtbOutput), true);
            Regex r = new Regex(@"[\40,\n,\r,\f]");
            string word = string.Empty;
            if (tp != null)
            {
                Int32 len = tp.GetTextInRun(LogicalDirection.Forward).Length;
                char c;
                for (int i = 0; i < len; i++)
                {
                    c = (char)tp.GetTextInRun(LogicalDirection.Forward)[i];
                    if (r.IsMatch(c.ToString()))
                        break;
                    word += c.ToString();
                }
                if (word.Length > 0)
                {
                    len = tp.GetTextInRun(LogicalDirection.Backward).Length;
                    for (int i = len - 1; 0 <= i; i--)
                    {
                        c = (char)tp.GetTextInRun(LogicalDirection.Backward)[i];
                        if (r.IsMatch(c.ToString()))
                            break;
                        word = c.ToString() + word; ;
                    }
                }
                if (word.Length > 0)
                {
                    User dcNick = Nicklist.GetbyNick(word);
                    if (dcNick != null)
                    {
                        this.rtbOutput.Cursor = System.Windows.Input.Cursors.Hand;
                    }
                    else
                    {
                        this.rtbOutput.Cursor = System.Windows.Input.Cursors.Arrow;
                    }
                }
                else
                {
                    this.rtbOutput.Cursor = System.Windows.Input.Cursors.Arrow;
                }
            }
            else
            {
                this.rtbOutput.Cursor = System.Windows.Input.Cursors.Arrow;
            }
        }

        void rtbOutput_MouseDown(object sender, System.Windows.RoutedEventArgs e)
        {
            Regex r = new Regex(@"[\40,\n,\r,\f]");
            TextPointer tp = rtbOutput.GetPositionFromPoint(System.Windows.Input.Mouse.GetPosition(this.rtbOutput), true);
            if (PreviewMouseDownArgs.ChangedButton == System.Windows.Input.MouseButton.Left)
            {
                if (PreviewMouseDownArgs.ClickCount == 2)
                {
                    string word = string.Empty;
                    Int32 len = tp.GetTextInRun(LogicalDirection.Forward).Length;
                    char c;
                    for (int i = 0; i < len; i++)
                    {
                        c = (char)tp.GetTextInRun(LogicalDirection.Forward)[i];
                        if (r.IsMatch(c.ToString()))
                            break;
                        word += c.ToString();
                    }
                    if (word.Length > 0)
                    {
                        len = tp.GetTextInRun(LogicalDirection.Backward).Length;
                        for (int i = len - 1; 0 <= i; i--)
                        {
                            c = (char)tp.GetTextInRun(LogicalDirection.Backward)[i];
                            if (r.IsMatch(c.ToString()))
                                break;
                            word = c.ToString() + word; ;
                        }
                    }
                    if (word.Length > 0)
                    {
                        User dcNick = Nicklist.GetbyNick(word);
                        if (dcNick != null)
                        {
                            dclick = true;
                            CreateWhisper(word);
                        }
                        else
                        {
                            r = new Regex("(?:^|[\\s\\[\\]\\}\\{\\(\\)\\\'\\\"<>])((?:(?:https?|gopher|ftp|file|irc):\\/\\/|www\\.)[a-zA-Z0-9\\.\\-=;&%\\?]+(?:\\/?[a-zA-Z0-9\\.\\-=;&%\\?#]*)*)");
                            if (!r.IsMatch(word))
                            {
                                if (inchan)
                                {
                                    if (chanman == null)
                                        chanman = new ChannelManager(this);
                                    if (chanman.IsDisposed)
                                        chanman = new ChannelManager(this);
                                    SendData("PROP " + Channel + " *");
                                    chanman.ShowDialog();
                                }
                            }
                        }
                    }
                    else
                    {
                        r = new Regex("(?:^|[\\s\\[\\]\\}\\{\\(\\)\\\'\\\"<>])((?:(?:https?|gopher|ftp|file|irc):\\/\\/|www\\.)[a-zA-Z0-9\\.\\-=;&%\\?]+(?:\\/?[a-zA-Z0-9\\.\\-=;&%\\?]*)*)");
                        if (!r.IsMatch(word))
                        {
                            if (inchan)
                            {
                                if (chanman == null)
                                    chanman = new ChannelManager(this);
                                if (chanman.IsDisposed)
                                    chanman = new ChannelManager(this);
                                SendData("PROP " + Channel + " *");
                                chanman.ShowDialog();
                            }
                        }
                    }
                }
            }
            if (PreviewMouseDownArgs.ChangedButton == System.Windows.Input.MouseButton.Right)
            {
                if (rtbOutput.Selection.IsEmpty)
                {
                    string word = string.Empty;
                    Int32 len = tp.GetTextInRun(LogicalDirection.Forward).Length;
                    char c;
                    for (int i = 0; i < len; i++)
                    {
                        c = (char)tp.GetTextInRun(LogicalDirection.Forward)[i];
                        if (r.IsMatch(c.ToString()))
                            break;
                        word += c.ToString();
                    }
                    if (word.Length > 0)
                    {
                        len = tp.GetTextInRun(LogicalDirection.Backward).Length;
                        for (int i = len - 1; 0 <= i; i--)
                        {
                            c = (char)tp.GetTextInRun(LogicalDirection.Backward)[i];
                            if (r.IsMatch(c.ToString()))
                                break;
                            word = c.ToString() + word; ;
                        }
                    }
                    if (word.Length > 0)
                    {
                        snick = Nicklist.GetbyNick(word);
                        if (snick != null)
                        {
                            User me = Nicklist.GetbyNick(Nickname);
                            mnuNicklist.Enabled = true;
                            if (snick.Ignore)
                            {
                                ignoreToolStripMenuItem.Text = "Unignore";
                                ignoreToolStripMenuItem.Image = imgNicknames.Images[15];
                            }
                            else
                            {
                                ignoreToolStripMenuItem.Text = "Ignore";
                                ignoreToolStripMenuItem.Image = imgNicknames.Images[14];
                            }

                            ownerToolStripMenuItem.Enabled = true;
                            ownerAccessMenu.Enabled = true;
                            hostToolStripMenuItem.Enabled = true;
                            helpOpToolStripMenuItem.Enabled = true;
                            voiceToolStripMenuItem.Enabled = true;
                            kickToolStripMenuItem.Enabled = true;
                            banToolStripMenuItem.Enabled = true;
                            accessToolStripMenuItem.Enabled = true;
                            ignoreToolStripMenuItem.Enabled = true;
                            whisperToolStripMenuItem.Enabled = true;
                            denyAccessMenu.Enabled = true;

                            if (me.Order > 5)
                            {
                                ownerToolStripMenuItem.Enabled = false;
                                ownerAccessMenu.Enabled = false;
                            }
                            if (me.Order > 7)
                            {
                                hostToolStripMenuItem.Enabled = false;
                                helpOpToolStripMenuItem.Enabled = false;
                                voiceToolStripMenuItem.Enabled = false;
                                kickToolStripMenuItem.Enabled = false;
                                banToolStripMenuItem.Enabled = false;
                                accessToolStripMenuItem.Enabled = false;
                            }
                            if (me == snick)
                            {
                                ignoreToolStripMenuItem.Enabled = false;
                                whisperToolStripMenuItem.Enabled = false;
                                kickToolStripMenuItem.Enabled = false;
                                banToolStripMenuItem.Enabled = false;
                                denyAccessMenu.Enabled = false;
                            }
                            this.mnuNicklist.Show(Cursor.Position.X, Cursor.Position.Y);
                        }
                    }
                }
                else
                {
                    //
                }
            }
        }

        void rtbOutput_PreviewMouseDown(object sender, System.Windows.Input.MouseButtonEventArgs e)
        {
            PreviewMouseDownArgs = e;
        }

        public bool IsConnected
        {
            get
            {
                return Buzzen.IsConnected;
            }
        }

        public bool InChannel
        {
            get
            {
                return inchan;
            }
        }

        public bool IsOwner
        {
            get
            {
                if (!Buzzen.IsConnected)
                    return false;
                if (!inchan)
                    return false;
                User me = Nicklist.GetbyNick(Nickname);
                if (me == null)
                    return false;
                if (me.Owner || me.Staff || me.Superowner)
                    return true;
                return false;
            }
        }

        public bool IsHost
        {
            get
            {
                if (!Buzzen.IsConnected)
                    return false;
                if (!inchan)
                    return false;
                User me = Nicklist.GetbyNick(Nickname);
                if (me == null)
                    return false;
                if (me.Host || me.Owner || me.Staff || me.Superowner)
                    return true;
                return false;
            }
        }

        public float GetFontSize()
        {
            return (float)rtbOutput.FontSize;
        }

        #region Control Events
        void TrayIcon_DoubleClick(object sender, EventArgs e)
        {
            this.TrayIcon.Visible = false;
            this.Show();
            WindowState = FormWindowState.Normal;
            if (alist != null)
                if (!alist.IsDisposed)
                    alist.Show();

            foreach (WhisperForm wf in Whispers)
            {
                wf.Show();
            }
        }

        void Client_Resize(object sender, EventArgs e)
        {
            if (FormWindowState.Minimized == WindowState)
            {
                foreach (WhisperForm wf in Whispers)
                {
                    wf.Hide();
                }
                if (alist != null)
                    if (!alist.IsDisposed)
                        alist.Hide();
                this.Hide();
                TrayIcon.Visible = true;
            }
        }

        void lvwNicknames_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (lvwNicknames.SelectedItems.Count < 1)
                mnuNicklist.Enabled = false;
            else
                mnuNicklist.Enabled = true;
        }

        void lvwNicknames_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                if (lvwNicknames.SelectedItems.Count > 0)
                {
                    snick = (User)lvwNicknames.SelectedItems[0].Tag;
                    if (Nickname != snick.Nickname)
                    {
                        dclick = true;
                        CreateWhisper(snick.Nickname);
                    }
                }
            }
        }

        void lvwNicknames_MouseClick(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                if (lvwNicknames.SelectedItems.Count > 0)
                {
                    User me = Nicklist.GetbyNick(Nickname);
                    snick = (User)lvwNicknames.SelectedItems[0].Tag;
                    if (snick.Ignore)
                    {
                        ignoreToolStripMenuItem.Text = "Unignore";
                        ignoreToolStripMenuItem.Image = imgNicknames.Images[15];
                    }
                    else
                    {
                        ignoreToolStripMenuItem.Text = "Ignore";
                        ignoreToolStripMenuItem.Image = imgNicknames.Images[14];
                    }

                    ownerToolStripMenuItem.Enabled = true;
                    ownerAccessMenu.Enabled = true;
                    hostToolStripMenuItem.Enabled = true;
                    helpOpToolStripMenuItem.Enabled = true;
                    voiceToolStripMenuItem.Enabled = true;
                    kickToolStripMenuItem.Enabled = true;
                    banToolStripMenuItem.Enabled = true;
                    accessToolStripMenuItem.Enabled = true;
                    ignoreToolStripMenuItem.Enabled = true;
                    whisperToolStripMenuItem.Enabled = true;
                    denyAccessMenu.Enabled = true;

                    if (me.Order > 5)
                    {
                        ownerToolStripMenuItem.Enabled = false;
                        ownerAccessMenu.Enabled = false;
                    }
                    if (me.Order > 7)
                    {
                        hostToolStripMenuItem.Enabled = false;
                        helpOpToolStripMenuItem.Enabled = false;
                        voiceToolStripMenuItem.Enabled = false;
                        kickToolStripMenuItem.Enabled = false;
                        banToolStripMenuItem.Enabled = false;
                        accessToolStripMenuItem.Enabled = false;
                    }
                    if (me == snick)
                    {
                        ignoreToolStripMenuItem.Enabled = false;
                        whisperToolStripMenuItem.Enabled = false;
                        kickToolStripMenuItem.Enabled = false;
                        banToolStripMenuItem.Enabled = false;
                        denyAccessMenu.Enabled = false;
                    }
                }
            }
        }

        void rtbInput_TextChanged(object sender, EventArgs e)
        {
            if (rtbInput.Text.IndexOf('\n') > -1)
            {
                string[] line = rtbInput.Text.Replace("\r", "").Split(new string[] { "\n" }, StringSplitOptions.RemoveEmptyEntries);
                for (int i = 0; i < line.Length; i++)
                    Execute(line[i]);
                rtbInput.Clear();
            }
        }

        void rtbInput_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.Control)
                if (e.KeyCode == Keys.V)
                {
                    rtbInput.Paste(DataFormats.GetFormat(DataFormats.Text));
                    e.Handled = true;
                }
            if (e.Shift)
                if (e.KeyCode == Keys.Insert)
                {
                    rtbInput.Paste(DataFormats.GetFormat(DataFormats.Text));
                    e.Handled = true;
                }
            if (e.KeyCode == Keys.Enter)
            {
                string[] line = rtbInput.Text.Replace("\r", "").Split(new string[] { "\n" }, StringSplitOptions.RemoveEmptyEntries);
                for (int i = 0; i < line.Length; i++)
                    Execute(line[i]);
                rtbInput.Clear();
                e.Handled = true;
            }
        }
        #endregion

        #region Primary Methods
        private void Connect()
        {
            if (Buzzen.IsConnected)
                Buzzen.Close();
            if ((Email == String.Empty) | (Password == String.Empty))
            {
                if (setup == null)
                {
                    setup = new Setup(this);
                }
                setup.ShowDialog();
            }
            else
            {
                moderated = false;
                try
                {
                    SetTitle("DragonChat - Attempting connection to irc.buzzen.net");
                    SetText("Attempting connection to irc.buzzen.net", System.Windows.Media.Colors.Navy, false, false);
                    Buzzen.Connect();
                    connectToolStripMenuItem.Enabled = false;
                    disconnectToolStripMenuItem.Enabled = true;
                }
                catch (SocketException ex)
                {
                    ReconnectAttempts++;
                    SetText("[" + ex.ErrorCode.ToString() + "] " + ex.Message, System.Windows.Media.Colors.Gray, false, false);
                    if (AttemptReconnect)
                    {
                        if (ReconnectAttempts > 3)
                        {
                            SetText("Attempt to reconnect failed.", System.Windows.Media.Colors.Gray, false, false);
                            AttemptReconnect = false;
                        }
                        else
                            Reconnect();
                    }
                }
            }
        }

        private void ToggleConnect(bool isconn)
        {
            if (this.InvokeRequired)
                this.Invoke(new BoolCallBack(this.ToggleConnect), new object[] { isconn });
            else
            {
                if (!isconn)
                {
                    disconnectToolStripMenuItem.Enabled = false;
                    connectToolStripMenuItem.Enabled = true;
                }
                else
                {
                    disconnectToolStripMenuItem.Enabled = true;
                    connectToolStripMenuItem.Enabled = false;
                }
            }
        }

        private void Reconnect()
        {
            if (this.InvokeRequired)
                this.Invoke(new CmdCallback(this.Reconnect));
            else
            {
                ReconnectAttempts++;
                ReconnectTimer.Interval = 5000;
                ReconnectTimer.Enabled = true;
            }
        }

        private void Disconnect()
        {
            if (this.InvokeRequired)
                this.Invoke(new CmdCallback(this.Disconnect));
            else
            {
                ClearNames();
                Buzzen.Close();
                disconnectToolStripMenuItem.Enabled = false;
                connectToolStripMenuItem.Enabled = true;
            }
        }

        public void SetText(string text)
        {
            if (rtbOutput.Dispatcher.CheckAccess())
            {
                if ((new TextRange(rtbOutput.Document.ContentStart, rtbOutput.Document.ContentEnd)).Text.Length > 2)
                {
                    new LineBreak(rtbOutput.Document.ContentEnd);
                }
                rtbOutput.Append(text);
                if (!rtbOutput.IsFocused)
                    rtbOutput.ScrollToEnd();
            }
            else
            {
                rtbOutput.Dispatcher.Invoke(new SetTextCallback(this.SetText), new object[] { text });
            }
        }

        public void SetText(string text, System.Windows.Media.Color color, bool bold, bool italic)
        {
            if (rtbOutput.Dispatcher.CheckAccess())
            {
                StringBuilder sb = new StringBuilder();
                if ((new TextRange(rtbOutput.Document.ContentStart, rtbOutput.Document.ContentEnd)).Text.Length > 2)
                {
                    new LineBreak(rtbOutput.Document.ContentEnd);
                }
                sb.Append(MakeStyle(rtbOutput.FontFamily.Source, color, bold, italic));
                sb.Append(text);
                sb.Append("[/style]");
                rtbOutput.Append(sb.ToString());
                if (!rtbOutput.IsFocused)
                    rtbOutput.ScrollToEnd();
            }
            else
            {
                rtbOutput.Dispatcher.Invoke(new SetColorTextCallback(this.SetText), new object[] { text, color, bold, italic });
            }
        }

        private void ShowTopic(string text)
        {
            if (!rtbOutput.Dispatcher.CheckAccess())
                this.Invoke(new SetTextCallback(this.ShowTopic), new object[] { text });
            else
            {
                StringBuilder sb = new StringBuilder();
                if ((new TextRange(rtbOutput.Document.ContentStart, rtbOutput.Document.ContentEnd)).Text.Length > 2)
                {
                    new LineBreak(rtbOutput.Document.ContentEnd);
                }
                sb.Append(text.Replace("[br]", "\f"));
                rtbOutput.Append(sb.ToString());

                if (!rtbOutput.IsFocused)
                    rtbOutput.ScrollToEnd();
            }
        }

        private void SetTitle()
        {
            if (this.InvokeRequired)
                this.Invoke(new CmdCallback(this.SetTitle));
            else
                this.Text = "DragonChat - " + Channel + " [" + Nicklist.Count.ToString() + "] [" + Modes + "]: " + StripStyle(Topic);
        }

        private void SetTitle(string text)
        {
            if (this.InvokeRequired)
            {
                SetTextCallback d = new SetTextCallback(SetTitle);
                this.Invoke(d, new object[] { text });
            } 
            else
            {
                this.Text = text;
            }
        }

        private void ClearScreen()
        {
            if (rtbOutput.Dispatcher.CheckAccess())
            {
                rtbOutput.Document.Blocks.Clear();
            }
            else
            {
                rtbOutput.Dispatcher.Invoke(new CmdCallback(this.ClearScreen));
            }
        }
        #endregion

        #region Execute Input Method
        private void Execute(string command)
        {
            if (!rtbOutput.Dispatcher.CheckAccess())
                this.Invoke(new SetTextCallback(this.Execute), new object[] { command });
            else
            {
                int i = 0;
                StringBuilder sb = new StringBuilder();
                foreach (Char c in command.ToCharArray())
                {
                    if (c == '/')
                        i++;
                    else
                        break;
                }
                if (i == 0)
                {
                    if (Buzzen.IsConnected)
                    {
                        command = GetStyle() + command + "[/style]";
                        SendData("PRIVMSG " + Channel + " :" + command);
                        if ((new TextRange(rtbOutput.Document.ContentStart, rtbOutput.Document.ContentEnd)).Text.Length > 2)
                        {
                            new LineBreak(rtbOutput.Document.ContentEnd);
                        }
                        sb.Append(MakeStyle(rtbOutput.FontFamily.Source, System.Windows.Media.Colors.Black, true, false));
                        sb.Append(Nickname);
                        sb.Append("[/style] : ");
                        sb.Append(command);
                        rtbOutput.Append(sb.ToString());

                        if (!rtbOutput.IsFocused)
                            rtbOutput.ScrollToEnd();
                    }
                    else
                        SetText("You must be connected to perform that action.", System.Windows.Media.Colors.Red, false, false);
                }
                else
                {
                    string[] words = command.Substring(i).Split(new string[] { " " }, StringSplitOptions.None);
                    switch (words[0].ToUpper())
                    {
                        case "ME":
                            if (Buzzen.IsConnected)
                            {
                                string input = string.Join(" ", words, 1, words.Length - 1);
                                SendData("PRIVMSG " + Channel + " :\u0001ACTION " + input + "\u0001");
                                if ((new TextRange(rtbOutput.Document.ContentStart, rtbOutput.Document.ContentEnd)).Text.Length > 2)
                                {
                                    new LineBreak(rtbOutput.Document.ContentEnd);
                                }
                                if (ShowTimeStamp)
                                {
                                    sb.Append(DateTime.Now.ToString("[HH:mm:ss] "));
                                }
                                sb.Append(MakeStyle(rtbOutput.FontFamily.Source, System.Windows.Media.Colors.Purple, false, true));
                                sb.Append(Nickname);
                                sb.Append(" ");
                                sb.Append(input);
                                sb.Append("[/style]");
                                rtbOutput.Append(sb.ToString());

                                if (!rtbOutput.IsFocused)
                                    rtbOutput.ScrollToEnd();
                            }
                            else
                                SetText("You must be connected to perform that action.", System.Windows.Media.Colors.Red, false, false);
                            break;
                        case "CLEAR":
                            ClearScreen();
                            break;
                        case "HOP":
                            if (Buzzen.IsConnected)
                            {
                                if (inchan)
                                    SendData("PART " + Channel + "\r\nJOIN " + Channel);
                            }
                            else
                                SetText("You must be connected to perform that action.", System.Windows.Media.Colors.Red, false, false);
                            break;
                        case "JOIN":
                            if (Buzzen.IsConnected)
                            {
                                Regex r = new Regex(@"^%#[^\s,%,#,\54]+$");
                                if (r.IsMatch(words[1]))
                                {
                                    if (words[1] != Channel)
                                        if (inchan)
                                            SendData("PART " + Channel);
                                    Channel = words[1];
                                    try
                                    {
                                        string ini = Application.StartupPath + @"\DragonChat.ini";
                                        IniFile iFile = new IniFile(ini);
                                        iFile.WriteValue("Settings", "Channel", Channel);
                                    }
                                    catch (Exception)
                                    {
                                        // do nothing
                                    }
                                    finally
                                    {
                                        SendData(command.Substring(i));
                                    }
                                }
                                else
                                    SetText("Invalid channel name: " + words[1], System.Windows.Media.Colors.Red, false, false);
                            }
                            else
                                SetText("You must be connected to perform that action.", System.Windows.Media.Colors.Red, false, false);
                            break;
                        case "PART":
                            break;
                        case "QUIT":
                            break;
                        case "KILL":
                            break;
                        case "KICK":
                            break;
                        case "NAMES":
                            break;
                        case "PRIVMSG":
                            break;
                        case "WHISPER":
                            break;
                        case "NOTICE":
                            break;
                        case "PING":
                            break;
                        case "PONG":
                            break;
                        case "AWAY":
                            if (Buzzen.IsConnected)
                            {
                                if (inchan)
                                {
                                    string input = string.Join(" ", words, 1, words.Length - 1);
                                    SendData("AWAY " + Channel + " " + input);
                                }
                            }
                            else
                                SetText("You must be connected to perform that action.", System.Windows.Media.Colors.Red, false, false);
                            break;
                        case "BACK":
                            if (Buzzen.IsConnected)
                            {
                                if (inchan)
                                    SendData("AWAY " + Channel);
                            }
                            else
                                SetText("You must be connected to perform that action.", System.Windows.Media.Colors.Red, false, false);
                            break;
                        default:
                            if (Buzzen.IsConnected)
                                SendData(command.Substring(i));
                            else
                                SetText("You must be connected to perform that action.", System.Windows.Media.Colors.Red, false, false);
                            break;
                    }
                }
            }
        }
        #endregion

        #region Nicklist Methods
        private void ClearNames()
        {
            if (lvwNicknames.InvokeRequired)
                this.Invoke(new CmdCallback(this.ClearNames));
            else
            {
                lvwNicknames.Items.Clear();
                Nicklist.Clear();
            }
        }

        private System.Windows.Forms.ListViewItem GetItem(string nick)
        {
            if (lvwNicknames.InvokeRequired)
                return (System.Windows.Forms.ListViewItem)this.Invoke(new GetItemCallback(this.GetItem), new object[] { nick });
            else
            {
                System.Windows.Forms.ListViewItem item = null;
                foreach (System.Windows.Forms.ListViewItem i in lvwNicknames.Items)
                    if (((User)i.Tag).Nickname == nick)
                        item = i;
                return item;
            }
        }

        private void ChangeIcon(System.Windows.Forms.ListViewItem item, int image)
        {
            if (lvwNicknames.InvokeRequired)
                this.Invoke(new ChangeIconCallback(this.ChangeIcon), new object[] { item, image });
            else
            {
                item.ImageIndex = image;
                if (image == 12)
                    item.ForeColor = System.Drawing.Color.Gray;
                else
                    item.ForeColor = System.Drawing.Color.Navy;

                User user = (User)item.Tag;

                StringBuilder sb = new StringBuilder(user.Nickname);

                switch (user.Order)
                {
                    case 1:
                        sb.Append(" (Staff)");
                        break;
                    case 2:
                    case 3:
                        sb.Append(" (Superowner)");
                        break;
                    case 4:
                    case 5:
                        sb.Append(" (Owner)");
                        break;
                    case 6:
                    case 7:
                        sb.Append(" (Host)");
                        break;
                    case 8:
                    case 9:
                        sb.Append(" (HelpOp)");
                        break;
                    default:
                        break;
                }

                item.Text = sb.ToString();
                lvwNicknames.Sort();
            }
        }

        private void ChangeName(System.Windows.Forms.ListViewItem item, string newnick)
        {
            if (lvwNicknames.InvokeRequired)
                this.Invoke(new ChangeNameCallback(this.ChangeName), new object[] { item, newnick });
            else
            {
                User user = (User)item.Tag;
                user.Nickname = newnick;
                
                StringBuilder sb = new StringBuilder(newnick);

                switch (user.Order)
                {
                    case 1:
                        sb.Append(" (Staff)");
                        break;
                    case 2:
                    case 3:
                        sb.Append(" (Superowner)");
                        break;
                    case 4:
                    case 5:
                        sb.Append(" (Owner)");
                        break;
                    case 6:
                    case 7:
                        sb.Append(" (Host)");
                        break;
                    case 8:
                    case 9:
                        sb.Append(" (HelpOp)");
                        break;
                    default:
                        break;
                }

                item.Text = sb.ToString();
                lvwNicknames.Sort();
            }
        }

        private void AddName(User user)
        {
            if (lvwNicknames.InvokeRequired)
            {
                AddUserCallback d = new AddUserCallback(this.AddName);
                this.Invoke(d, new object[] { user });
            }
            else
            {
                Nicklist.Add(user, moderated);
                System.Drawing.Color color;
                StringBuilder sb = new StringBuilder(user.Nickname);

                switch (user.Order)
                {
                    case 1:
                        sb.Append(" (Staff)");
                        break;
                    case 2:
                    case 3:
                        sb.Append(" (Superowner)");
                        break;
                    case 4:
                    case 5:
                        sb.Append(" (Owner)");
                        break;
                    case 6:
                    case 7:
                        sb.Append(" (Host)");
                        break;
                    case 8:
                    case 9:
                        sb.Append(" (HelpOp)");
                        break;
                    default:
                        break;
                }

                if (user.ImageIndex == 12)
                    color = System.Drawing.Color.Gray;
                else
                    color = System.Drawing.Color.Navy;

                System.Windows.Forms.ListViewItem item = new System.Windows.Forms.ListViewItem(new string[] { sb.ToString() }, user.ImageIndex, color, lvwNicknames.BackColor, lvwNicknames.Font);
                item.Tag = user;
                item.ToolTipText = sb.ToString();
                lvwNicknames.Items.Add(item);
                SetTitle();
            }
        }

        private void DelName(string value)
        {
            if (lvwNicknames.InvokeRequired)
            {
                SetTextCallback d = new SetTextCallback(this.DelName);
                this.Invoke(d, new object[] { value });
            }
            else
            {
                User user = Nicklist.GetbyNick(value);
                System.Windows.Forms.ListViewItem item = null;

                foreach (System.Windows.Forms.ListViewItem i in lvwNicknames.Items)
                    if ((User)i.Tag == user)
                        item = i;
                if (user != null)
                    Nicklist.Remove(user);
                if (item != null)
                    item.Remove();
                SetTitle();
            }
        }

        private void UpdateList()
        {
            if (lvwNicknames.InvokeRequired)
            {
                CmdCallback d = new CmdCallback(this.UpdateList);
                this.Invoke(d);
            }
            else
            {
                lvwNicknames.BeginUpdate();
                lvwNicknames.Items.Clear();
                User[] Users = Nicklist.GetUsers();
                System.Windows.Forms.ListViewItem item;
                System.Drawing.Color color;
                StringBuilder sb = new StringBuilder();
                for (int i = 0; i < Users.Length; i++)
                {
                    sb.Append(Users[i].Nickname);

                    switch (Users[i].Order)
                    {
                        case 1:
                            sb.Append(" (Staff)");
                            break;
                        case 2:
                        case 3:
                            sb.Append(" (Superowner)");
                            break;
                        case 4:
                        case 5:
                            sb.Append(" (Owner)");
                            break;
                        case 6:
                        case 7:
                            sb.Append(" (Host)");
                            break;
                        case 8:
                        case 9:
                            sb.Append(" (HelpOp)");
                            break;
                        default:
                            break;
                    }

                    if (Users[i].ImageIndex == 12)
                        color = System.Drawing.Color.Gray;
                    else
                        color = System.Drawing.Color.Navy;
                    item = new System.Windows.Forms.ListViewItem(new string[] { sb.ToString() }, Users[i].ImageIndex, color, lvwNicknames.BackColor, lvwNicknames.Font);
                    item.ToolTipText = sb.ToString();
                    item.Tag = Users[i];
                    lvwNicknames.Items.Add(item);
                    sb.Remove(0, sb.Length);
                }
                SetTitle();
                lvwNicknames.EndUpdate();
            }
        }
        #endregion

        #region Whispers
        private WhisperForm CreateWhisper(string nickname)
        {
            if (this.InvokeRequired)
            {
                CreateWhisperCallback d = new CreateWhisperCallback(this.CreateWhisper);
                return (WhisperForm)this.Invoke(d, new object[] { nickname });
            }
            else
            {
                WhisperForm whisper = Whispers.Get(nickname);
                if (whisper == null)
                {
                    whisper = new WhisperForm(nickname);
                    if (!dclick)
                    {
                        whisper.WindowState = FormWindowState.Minimized;
                        dclick = false;
                    }
                    whisper.OnWhisperInput += new WhisperForm.WhisperInputCallback(OnWhisperInput);
                    whisper.OnWhisperClose += new WhisperForm.WhisperCloseCallback(OnWhisperClose);
                    whisper.Show();
                    if (FormWindowState.Minimized == WindowState)
                        whisper.Hide();
                    Whispers.Add(whisper);
                }
                return whisper;
            }
        }

        private void OnWhisperInput(WhisperForm whisper, string text)
        {
            int i = 0;
            foreach (Char c in text.ToCharArray())
            {
                if (c == '/')
                    i++;
                else
                    break;
            }
            if (i == 0)
            {
                SendData("WHISPER " + Channel + " " + whisper.Nickname + " :" + text);
                text = StripmIRC(text);
                whisper.DisplayWhisper(Nickname, text);
            }
            else
            {
                string[] words = text.Substring(i).Split(new string[] { " " }, StringSplitOptions.None);
                switch (words[0].ToUpper())
                {
                    case "ME":
                        string input = StripmIRC(string.Join(" ", words, 1, words.Length - 1));
                        SendData("WHISPER " + Channel + " " + whisper.Nickname + " :\u0001ACTION " + input + "\u0001");
                        whisper.SetText(Nickname + " " + input, System.Windows.Media.Colors.Purple, false, true);
                        break;
                    case "CLEAR":
                        whisper.Clear();
                        break;
                    default:
                        whisper.SetText("Unsupported whisper command: " + words[0].ToUpper(), System.Windows.Media.Colors.Gray, false, false); 
                        break;
                }
            }
        }

        private void OnWhisperClose(WhisperForm whisper)
        {
            Whispers.Remove(whisper);
        }
        #endregion

        #region Chat Events
        private void OnJoin(string nick, string addr, string pdata)
        {
            if (nick != Nickname)
            {
                DelName(nick);
                User user = Nicklist.GetbyNick(nick);
                if (user == null)
                {
                    user = new User(nick, addr, pdata);
                    AddName(user);
                }
                SetText("\t " + nick + " has joined the conversation.", System.Windows.Media.Colors.Gray, false, false);
            }
            else
            {
                string hash = MD5Hash(Channel);
                Topic = string.Empty;
                OnjoinMessage = string.Empty;
                SetText("\t You have entered " + Channel, System.Windows.Media.Colors.Gray, false, false);
                inchan = true;
                IniFile iFile = new IniFile(Application.StartupPath + @"\History.ini");
                iFile.WriteValue(hash, "Name", Channel);
                iFile.WriteValue(hash, "LastJoin", GetTimeStamp());
                Ownerkey = iFile.GetString(hash, "Ownerkey", string.Empty);
                Hostkey = iFile.GetString(hash, "Hostkey", string.Empty);
                Category = iFile.GetString(hash, "Category", string.Empty);
                SendData("MODE " + Channel);
                SendData("WHO " + Channel);
                iFile = null;
            }
        }

        private void OnPart(string nick, string addr)
        {
            if (nick == Nickname)
            {
                SetText("\t You have left the conversation.", System.Windows.Media.Colors.Gray, false, false);
                inchan = false;
                ClearNames();
            }
            else
            {
                DelName(nick);
                SetText("\t " + nick + " has left the conversation.", System.Windows.Media.Colors.Gray, false, false);
            }
        }

        private void OnKick(string nick, string addr, string knick, string text)
        {
            if (knick == Nickname)
            {
                inchan = false;
                moderated = false;
                ClearNames();
                if (rejoin)
                    SendData("JOIN " + Channel);
            }
            else
                DelName(knick);
            SetText(nick + " kicked " + knick + " from " + Channel + " : " + text, System.Windows.Media.Colors.Red, true, false);
        }

        private void OnKill(string nick, string addr, string knick, string text)
        {
            if (knick == Nickname)
            {
                inchan = false;
                moderated = false;
                ClearNames();
            }
            else
                DelName(knick);
            SetText(nick + " kicked " + knick + " from the server.", System.Windows.Media.Colors.Red, true, false);
        }

        private void OnMode(string nick, string addr, string mode, string value)
        {
            int i = 0;
            bool setmode = false;
            bool domoderate = moderated;
            User user;
            System.Windows.Forms.ListViewItem item;
            
            string[] m = Modes.Split(new string[] { " " }, StringSplitOptions.None);
            string current_modes = m[0].Replace("+","");
            string limit = "0";
            if (m.Length > 1)
                limit = m[1];
            string rkey = string.Empty;
            if (m.Length > 2)
                rkey = m[2];

            string[] mode_params = value.Split(new string[] { " " }, StringSplitOptions.None);
            SetText("\t " + nick + " sets mode " + mode + " " + value, System.Windows.Media.Colors.Gray, false, false);
            foreach (Char c in mode.ToCharArray())
            {
                switch (c)
                {
                    case '-':
                        setmode = false;
                        break;
                    case '+':
                        setmode = true;
                        break;
                    case 'Q':
                        user = Nicklist.GetbyNick(mode_params[i]);
                        if (user != null)
                        {
                            user.Superowner = setmode;
                            item = GetItem(mode_params[i]);
                            if (item != null)
                                ChangeIcon(item, user.ImageIndex);
                        }
                        i++;
                        break;
                    case 'q':
                        user = Nicklist.GetbyNick(mode_params[i]);
                        if (user != null)
                        {
                            user.Owner = setmode;
                            item = GetItem(mode_params[i]);
                            if (item != null)
                                ChangeIcon(item, user.ImageIndex);
                        }
                        i++;
                        break;
                    case 'o':
                        user = Nicklist.GetbyNick(mode_params[i]);
                        if (user != null)
                        {
                            user.Host = setmode;
                            item = GetItem(mode_params[i]);
                            if (item != null)
                                ChangeIcon(item, user.ImageIndex);
                        }
                        i++;
                        break;
                    case 'h':
                        user = Nicklist.GetbyNick(mode_params[i]);
                        if (user != null)
                        {
                            user.HelpOp = setmode;
                            item = GetItem(mode_params[i]);
                            if (item != null)
                                ChangeIcon(item, user.ImageIndex);
                        }
                        i++;
                        break;
                    case 'v':
                        user = Nicklist.GetbyNick(mode_params[i]);
                        if (user != null)
                        {
                            user.Voice = setmode;
                            if (moderated)
                            {
                                item = GetItem(mode_params[i]);
                                if (item != null)
                                    ChangeIcon(item, user.ImageIndex);
                            }
                        }
                        i++;
                        break;
                    case 'm':
                        if (setmode)
                        {
                            if (current_modes.IndexOf(c) == -1)
                                current_modes += c.ToString();
                            domoderate = true;
                        }
                        else
                        {
                            current_modes = current_modes.Replace(c.ToString(), "");
                            domoderate = false;
                        }
                        break;
                    case 'k':
                        if (setmode)
                        {
                            if (current_modes.IndexOf(c) == -1)
                                current_modes += c.ToString();
                            rkey = mode_params[i];
                            i++;
                        }
                        else
                        {
                            current_modes = current_modes.Replace(c.ToString(), "");
                            rkey = string.Empty;
                        }
                        break;
                    case 'l':
                        if (setmode)
                        {
                            if (current_modes.IndexOf(c) == -1)
                                current_modes += c.ToString();
                            limit = mode_params[i];
                            i++;
                        }
                        break;
                    default:
                        if (setmode)
                        {
                            if (current_modes.IndexOf(c) == -1)
                                current_modes += c.ToString();
                        }
                        else
                        {
                            current_modes = current_modes.Replace(c.ToString(), "");
                        }
                        break;
                }
            }
            // Build Mode 
            if (current_modes.IndexOf('l') > -1)
                Modes = "l";
            current_modes = current_modes.Replace("l", "");
            if (current_modes.IndexOf('k') > -1)
                Modes += "k";
            current_modes = current_modes.Replace("k", "");
            char[] _c = current_modes.ToCharArray();
            Array.Sort(_c);
            current_modes = new string(_c);
            Modes = "+" + current_modes + Modes + " " + limit;
            if (rkey != string.Empty)
                Modes += " " + rkey;
            SetTitle();
            if (moderated != domoderate)
            {
                moderated = domoderate;
                Nicklist.SetModerated(moderated);
                UpdateList();
            }
        }

        private void OnText(string nick, string addr, string text)
        {
            if (!rtbOutput.Dispatcher.CheckAccess())
                rtbOutput.Dispatcher.Invoke(new OnTextCallback(this.OnText), new object[] { nick, addr, text });
            else
            {
                string chkval = StripmIRC(text);
                if (nick == Channel)
                {
                    OnjoinMessage = chkval;
                    SetText(StripStyle(chkval).Replace("&nbsp", " ").Replace("[br]", "\f"), System.Windows.Media.Colors.Green, true, false);
                }
                else
                {
                    Regex r = new Regex(@"^\u0001(\S+)(?:\s(.*)?)?\u0001$");
                    Match mr = r.Match(chkval);
                    if (mr.Success)
                    {
                        if (mr.Groups[1].Value.ToUpper() == "ACTION")
                            OnAction(nick, addr, mr.Groups[2].Value);
                        else
                            OnCTCPRequest(nick, addr, mr.Groups[1].Value);
                    }
                    else
                    {
                        if (text == "\u0002TIME\u0002")
                            OnCTCPRequest(nick, addr, "TIME");
                        else
                            if (!Nicklist.GetbyNick(nick).Ignore)
                            {
                                StringBuilder sb = new StringBuilder();
                                if (ShowTimeStamp)
                                {
                                    sb.Append(DateTime.Now.ToString("[HH:mm:ss] "));
                                }
                                if ((new TextRange(rtbOutput.Document.ContentStart, rtbOutput.Document.ContentEnd)).Text.Length > 2)
                                {
                                    new LineBreak(rtbOutput.Document.ContentEnd);
                                }
                                sb.Append(MakeStyle(rtbOutput.FontFamily.Source, System.Windows.Media.Colors.Navy, false, false));
                                sb.Append(nick);
                                sb.Append("[/style]");
                                sb.Append(" : ");
                                sb.Append(chkval);
                                rtbOutput.Append(sb.ToString());

                                if (!rtbOutput.IsFocused)
                                    rtbOutput.ScrollToEnd();
                            }
                    }
                }
            }
        }

        private void OnAction(string nick, string addr, string text)
        {
            if (!Nicklist.GetbyNick(nick).Ignore)
                SetText(nick + " " + text, System.Windows.Media.Colors.Purple, false, true);
        }

        private void OnNotice(string nick, string addr, string text)
        {
            string chkval = StripStyle(StripmIRC(text));
            Regex r = new Regex(@"^\u0001(\S+)(?:\s(.*)?)?\u0001$");
            Match m = r.Match(chkval);
            if (m.Success)
            {
                OnCTCPReply(nick, addr, m.Groups[1].Value, m.Groups[2].Value);
            }
            else
            {
                if (text.Length > 5)
                {
                    if (text.Substring(0, 6).ToUpper() == "\u0002TIME\u0002")
                        OnCTCPReply(nick, addr, "TIME", text.Substring(6));
                    else
                        if (!Nicklist.GetbyNick(nick).Ignore)
                            SetText("Notice(" + nick + "): " + chkval);
                }
                else
                    if (!Nicklist.GetbyNick(nick).Ignore)
                        SetText("Notice(" + nick + "): " + chkval);
            }
        }

        private void OnServerNotice(string text)
        {
            SetText(text, System.Windows.Media.Colors.SteelBlue, false, false);
            if (text.IndexOf("*** authentication failed:") > -1)
            {
                AttemptReconnect = false;
                Disconnect();
            }
        }

        private void OnCTCPRequest(string nick, string addr, string text)
        {
            if (text.ToUpper() == "TIME")
                if (!Nicklist.GetbyNick(nick).Ignore)
                    SetText("\t " + nick + " has requested your local time.", System.Windows.Media.Colors.Gray, false, false);
            if (text.ToUpper() == "VERSION")
                if (!Nicklist.GetbyNick(nick).Ignore)
                    SetText("\t " + nick + " has requested your version information.", System.Windows.Media.Colors.Gray, false, false);
        }

        private void OnCTCPReply(string nick, string addr, string type, string text)
        {
            if (type.ToUpper() == "TIME")
                if (!Nicklist.GetbyNick(nick).Ignore)
                    SetText(nick + "'s local time is " + text, System.Windows.Media.Colors.Gray, false, false);
            if (type.ToUpper() == "VERSION")
                if (!Nicklist.GetbyNick(nick).Ignore)
                    SetText(nick + "'s version information is " + text, System.Windows.Media.Colors.Gray, false, false);
        }

        private void OnProp(string nick, string addr, string type, string text)
        {
            IniFile iFile = new IniFile(Application.StartupPath + "\\History.ini");
            switch (type.ToUpper())
            {
                case "TOPIC":
                    Topic = StripmIRC(text);
                    SetTitle();
                    SetText("\t [style co:#236B8E;]" + nick + "[/style] has changed the [style co:#236B8E;]topic[/style] for [style co:#236B8E;]" + Channel + "[/style].", System.Windows.Media.Colors.Gray, false, false);
                    ShowTopic(text);
                    break;
                case "OWNERKEY":
                    iFile.WriteValue(MD5Hash(Channel), "Name", Channel);
                    iFile.WriteValue(MD5Hash(Channel), "Ownerkey", Decode(text));
                    Ownerkey = Decode(text);
                    break;
                case "HOSTKEY":
                    iFile.WriteValue(MD5Hash(Channel), "Name", Channel);
                    iFile.WriteValue(MD5Hash(Channel), "Hostkey", Decode(text));
                    Hostkey = Decode(text);
                    break;
                case "CATEGORY":
                    iFile.WriteValue(MD5Hash(Channel), "Name", Channel);
                    iFile.WriteValue(MD5Hash(Channel), "Category", text);
                    Category = text;
                    break;
                case "ONJOIN":
                    OnjoinMessage = StripmIRC(text);
                    break;
                case "ONPART":
                    break;
                case "LOCKED":
                    break;
                case "OID": // Umm...
                    break;
                case "NAME":
                    iFile.WriteValue(MD5Hash(Channel), "Name", text);
                    break;
                case "CREATION":
                    iFile.WriteValue(MD5Hash(Channel), "Creation", text);
                    break;
                case "LANGUAGE":
                    iFile.WriteValue(MD5Hash(Channel), "Language", text);
                    break;
                case "LANGUAGE2": // Obsolete
                    break;
                case "MEMBERKEY": // Not used
                    break;
                case "PICS": // Not used
                    break;
                default:
                    SetText("\t [style co:#236B8E;]" + nick + "[/style] has changed the [style co:#236B8E;]" + type + "[/style] property to: [style co:#000000;]" + text + "[/style]", System.Windows.Media.Colors.Gray, false, false);
                    break;
            }
        }

        private void OnNickChange(string nick, string addr, string newnick)
        {
            if (string.IsNullOrEmpty(nick))
            {
                if (!string.IsNullOrEmpty(Nickname))
                {
                    if (newnick.ToLower() == Nickname.ToLower())
                        Nickname = newnick;
                    if (Nickname != newnick)
                        SendData("NICK " + Nickname);
                }
                else
                {
                    Nickname = newnick;
                }
                SendData("USER " + Nickname + " " + Nickname + " " + Nickname + " :DragonChat");
            }
            else
            {
                if (nick == Nickname)
                    Nickname = newnick;
                WhisperForm whisper = Whispers.Get(nick);
                if (whisper != null)
                    whisper.Nickname = newnick;
                System.Windows.Forms.ListViewItem item = GetItem(nick);
                if (item != null)
                {
                    User user = (User)item.Tag;
                    ChangeName(item, newnick);
                    if (Nickname == newnick)
                    {
                        if (!user.Ignore)
                            SetText("\t You are now known as " + newnick, System.Windows.Media.Colors.Gray, false, false);
                    }
                    else
                        if (!user.Ignore)
                            SetText("\t " + nick + " is now known as " + newnick, System.Windows.Media.Colors.Gray, false, false);
                }
            }
        }

        private void OnWhisper(string nick, string addr, string text)
        {
            User user = Nicklist.GetbyNick(nick);
            if (!user.Ignore)
            {
                WhisperForm whisper = CreateWhisper(nick);
                whisper.DisplayWhisper(nick, text);
            }
        }

        private void OnQuery(string nick, string addr, string text)
        {
            string chkval = StripStyle(StripmIRC(text));
            Regex r = new Regex(@"^\u0001(\S+)(?:\s(.*)?)?\u0001$");
            Match m = r.Match(chkval);
            if (m.Success)
            {
                if (m.Groups[1].Value.ToUpper() == "ACTION")
                    OnWhisper(nick, addr, chkval);
                else
                    OnCTCPRequest(nick, addr, m.Groups[1].Value);
            }
            else
                OnWhisper(nick, addr, text);
        }

        private void OnNumeric(string data)
        {
            string[] words = data.Split(new string[] { " " }, StringSplitOptions.RemoveEmptyEntries);
            Regex r;
            Match mr;
            System.Windows.Forms.ListViewItem item;
            User user;
            AccessEntry ae;
            string comment = string.Empty;
            switch (words[1])
            {
                case "001":
                case "002":
                case "003":
                case "004":
                case "005":
                case "251":
                case "422":
                case "329": // :FlashIrcServer.1 329 Tewl %#!!~\bSoulFly\b~!! 1195078304
                    break;
                case "332":
                    r = new Regex(@"^:[^:]*:(.*)$");
                    mr = r.Match(data);
                    if (mr.Success)
                    {
                        Topic = StripmIRC(mr.Groups[1].Value);
                        ShowTopic(Topic);
                        SetTitle();
                    }
                    break;
                case "324":
                    if (words[4].IndexOf('m') > -1)
                        if (!moderated)
                        {
                            moderated = true;
                            Nicklist.SetModerated(moderated);
                            UpdateList();
                        }
                    Modes = string.Join(" ", words, 4, words.Length - 4);
                    SetTitle();
                    break;
                case "265":
                    if (Channel != string.Empty)
                        SendData("JOIN " + Channel);
                    break;
                case "315": // :FlashIrcServer.1 315 Tewl %#DevCafe :End of /WHO list.
                    break;
                case "352": // :FlashIrcServer.1 352 Tewl %#DevCafe 69d1fc78dbda242c43ad6590368912d4 BuzzenPassport FlashIrcServer.1 YAH00™ H' :0 anon
                    User usr = Nicklist.GetbyNick(words[7]);
                    if (usr == null)
                    {
                        User[] test = Nicklist.GetUsers();
                        foreach (User u in test)
                        {
                            SetText(words[7] + " " + u.Nickname + " " + u.ToString());
                        }
                    }
                    else
                        usr.Address = words[4] + "@" + words[5];
                    break;
                case "353":
                    r = new Regex(@"^:[^:]*:(.*)$");
                    mr = r.Match(data);
                    if (mr.Success)
                    {
                        string[] names = mr.Groups[1].Value.Split(new string[] { " " }, StringSplitOptions.RemoveEmptyEntries);
                        if (flushed)
                        {   // Clear collection if this is first 353 sent. 
                                Nicklist.Clear();
                                flushed = false;
                        }
                        for (int i = 0; i < names.Length; i++)
                        {
                            user = new User(names[i].Substring(5), names[i].Substring(0, 4));
                            Nicklist.Add(user, moderated);
                        }
                    }
                    break;
                case "366": // Update list and reset flushed to true 
                    UpdateList();
                    flushed = true;
                    break;
                case "433": //:FlashIrcServer.1 433 Tewlz Tewl :Nickname is already in use
                    Nickname = words[2];
                    SetText("*** Nickname: [style b;]" + words[3] + "[/style] is already in use.", System.Windows.Media.Colors.SteelBlue, false, false);
                    SetText("\t You are now known as " + words[2], System.Windows.Media.Colors.Gray, false, false);
                    break;
                case "821": // unaway
                    r = new Regex(@"^:([^\40]+)!([^\40]+@[^\40]+)\40(?:821)\40([^\40]+\40)?:(.*)?$");
                    mr = r.Match(data);
                    if (mr.Success)
                    {
                        item = GetItem(mr.Groups[1].Value);
                        if (item != null)
                        {
                            user = (User)item.Tag;
                            user.Away = false;
                            ChangeIcon(item, user.ImageIndex);
                            SetText("\t " + mr.Groups[1].Value + " is no longer marked as away.", System.Windows.Media.Colors.Gray, false, false);
                        }
                    }
                    else
                    {
                        SetText(data);
                    }
                    break;
                case "801": // :FlashIrcServer.1 801 Tewl!8ccd6e7d066dc6325428b0b9373e270c@BuzzenPassport %#!!~\bSoulFly\b~!! OWNER *!8ccd6e7d066dc6325428b0b9373e270c@BuzzenPassport 0 Tewl!8ccd6e7d066dc6325428b0b9373e270c@BuzzenPassport : Tewl
                    if (words.Length > 9)
                    {
                        comment = StripmIRC(string.Join(" ", words, 8, words.Length - 8)).Substring(1).Trim(); ;
                    }
                    else
                    {
                        comment = string.Empty;
                    }
                    ae = new AccessEntry(words[5], words[7], words[4], words[6], comment);
                    AccessEntries.Add(ae);
                    UpdateAccess();
                    SetText("\t Added [style co:#236B8E;]" + words[5] + "[/style] to the channel [style co:#236B8E;]" + words[4].ToLower() + "[/style] list.", System.Windows.Media.Colors.Gray, false, false);
                    break;
                case "802": // :FlashIrcServer.1 802 Tewl %#!!~\bSoulFly\b~!! OWNER *!8ccd6e7d066dc6325428b0b9373e270c@BuzzenPassport
                    bool removed = AccessEntries.RemoveEntry(words[5], words[4]);
                    if (removed)
                        UpdateAccess();
                    SetText("\t Removed [style co:#236B8E;]" + words[5] + "[/style] from the channel [style co:#236B8E;]" + words[4].ToLower() + "[/style] list.", System.Windows.Media.Colors.Gray, false, false);
                    break;
                case "803":
                    SetText("\t Retrieving access list.", System.Windows.Media.Colors.SteelBlue, false, false);
                    AccessEntries.Clear();
                    break;
                case "804":
                    if (words.Length > 9)
                    {
                        comment = StripmIRC(string.Join(" ", words, 8, words.Length - 8)).Substring(1).Trim(); ;
                    }
                    else
                    {
                        comment = string.Empty;
                    }
                    ae = new AccessEntry(words[5], words[7], words[4], words[6], comment);
                    AccessEntries.Add(ae);
                    break;
                case "805":
                    UpdateAccess();
                    break;
                case "820": // :FlashIrcServer.1 820 Tewl %#Tewlzbox DENY :Clear
                    string list = words[4].ToLower();
                    if (list == "*")
                        list = "access";
                    else
                        list = "[style co:#236B8E;]" + list + "[/style]";
                    SetText("\t You have [style co:#236B8E;]cleared[/style] the channel " + list + " list.", System.Windows.Media.Colors.Gray, false, false);
                    SendData("ACCESS " + Channel + " LIST");
                    break;
                case "818":
                    IniFile iFile = new IniFile(Application.StartupPath + "\\History.ini");
                    string sval = string.Join(" ", words, 5, words.Length - 5).Substring(1);
                    switch (words[4].ToUpper())
                    {
                        case "TOPIC":
                            Topic = StripmIRC(sval);
                            SetTitle();
                            break;
                        case "OWNERKEY":
                            Ownerkey = Decode(sval);
                            iFile.WriteValue(MD5Hash(Channel), "Ownerkey", Decode(sval));
                            break;
                        case "HOSTKEY":
                            Hostkey = Decode(sval);
                            iFile.WriteValue(MD5Hash(Channel), "Hostkey", Decode(sval));
                            break;
                        case "CATEGORY":
                            Category = sval;
                            iFile.WriteValue(MD5Hash(Channel), "Category", sval);
                            break;
                        case "ONJOIN":
                            OnjoinMessage = StripmIRC(sval);
                            break;
                        case "ONPART":
                            break;
                        case "LOCKED":
                            break;
                        case "OID": // Umm...
                            break;
                        case "NAME":
                            iFile.WriteValue(MD5Hash(Channel), "Name", sval);
                            break;
                        case "CREATION":
                            iFile.WriteValue(MD5Hash(Channel), "Creation", sval);
                            break;
                        case "LANGUAGE":
                            iFile.WriteValue(MD5Hash(Channel), "Language", sval);
                            break;
                    }
                    break;
                case "819":
                    UpdateSettings();
                    break;
                case "822": // away
                    r = new Regex(@"^:([^\40]+)!([^\40]+@[^\40]+)\40(?:822)\40([^\40]+\40)?:(.*)?$");
                    mr = r.Match(data);
                    if (mr.Success)
                    {
                        item = GetItem(mr.Groups[1].Value);
                        if (item != null)
                        {
                            user = (User)item.Tag;
                            user.Away = true;
                            ChangeIcon(item, user.ImageIndex);
                            SetText("\t " + mr.Groups[1].Value + " has been marked as away.", System.Windows.Media.Colors.Gray, false, false);
                        }
                    }
                    else
                    {
                        SetText(data);
                    }
                    break;
                case "403": // :FlashIrcServer.1 403 Tewl[c] %#test :No such channel
                    r = new Regex(@"^%#[^\s,%,#,\54]+$");
                    if (r.IsMatch(words[3]))
                    {
                        SendData("CREATE " + words[3]);
                    }
                    else
                        SetText("\t Invalid channel name: " + words[3], System.Windows.Media.Colors.Red, false, false);
                    break;
                case "471": // :FlashIrcServer.1 471 Tewl[c] %#test :Cannot join channel (+l)
                    SetText("\t Unable to join channel. Room is full.", System.Windows.Media.Colors.Red, false, false);
                    break;
                case "473": // :FlashIrcServer.1 473 Tewl[c] %#test :Cannot join channel (+i)
                    SetText("\t Unable to join channel. Room is invite only.", System.Windows.Media.Colors.Red, false, false);
                    break;
                case "474": // :FlashIrcServer.1 474 Tewl[c] %#test :Cannot join channel (+b)
                    SetText("\t Unable to join channel. Your nickname/address is banned.", System.Windows.Media.Colors.Red, false, false);
                    break;
                case "475": // :FlashIrcServer.1 475 Tewl[c] %#test :Cannot join channel (+k)
                    SetText("\t Unable to join channel. Key required.", System.Windows.Media.Colors.Red, false, false);
                    break;
                default:
                    SetText("\t Numeric([style co:#236B8E;]" + words[1] + "[/style]): " + string.Join(" ",words,3,words.Length - 3), System.Windows.Media.Colors.Gray, false, false);
                    break;
            }
        }
        #endregion

        #region Parser
        private void Parse(string data)
        {
            string[] lines = data.Split(new string[] { "\r\n" }, StringSplitOptions.None);
            string[] words = null;
            string line;
            Regex r;
            Match rm;
            for (int i = 0; i < lines.Length; i++)
            {
                line = lines[i];
                words = line.Split(new string[] { " " }, StringSplitOptions.RemoveEmptyEntries);
                if (words.Length > 0)
                {
                    if (words[0] == "PING")
                        SendData("PONG " + words[1]);
                    else
                    {
                        if (IsNumeric(words[1]))
                            OnNumeric(line);
                        else
                        {
                            switch (words[1])
                            {
                                case "JOIN":
                                    r = new Regex(@"^:([^\40]+)!([^\40]+@[^\40]+)\40(?:JOIN)\40([A,U][U,F,M][N,P][V,N])");
                                    rm = r.Match(line);
                                    if (rm.Success)
                                    {
                                        if (rm.Groups[1].Value == Nickname)
                                            Channel = words[3].Remove(0, 1);
                                        OnJoin(rm.Groups[1].Value, rm.Groups[2].Value, rm.Groups[3].Value);
                                    }
                                    else
                                    {
                                        // Handle Staff join messages
                                        r = new Regex(@"^:([^\40]+)!([^\40]+@[^\40]+)\40(?:JOIN)");
                                        rm = r.Match(line);
                                        if (rm.Success)
                                        {
                                            if (rm.Groups[1].Value == Nickname)
                                                Channel = words[3].Remove(0, 1);
                                            OnJoin(rm.Groups[1].Value, rm.Groups[2].Value, "UUNN");
                                        }
                                    }
                                    break;
                                case "PART":
                                    r = new Regex(@"^:([^\40]+)!([^\40]+@[^\40]+)\40(?:PART)");
                                    rm = r.Match(line);
                                    if (rm.Success)
                                        OnPart(rm.Groups[1].Value, rm.Groups[2].Value);
                                    break;
                                case "KICK":
                                    r = new Regex(@"^:([^\40]+)!([^\40]+@[^\40]+)\40(?:KICK)\40:?[^\40]*\40([^\40]*)\40:(.*)?$");
                                    rm = r.Match(line);
                                    if (rm.Success)
                                        OnKick(rm.Groups[1].Value, rm.Groups[2].Value, rm.Groups[3].Value, rm.Groups[4].Value);
                                    break;
                                case "NOTICE":
                                    r = new Regex(@"^:([^\40]+)!([^\40]+@[^\40]+)\40(?:NOTICE)\40:?[^\40]*\40[^\40]*\40:(.*)?$");
                                    rm = r.Match(line);
                                    if (rm.Success)
                                        OnNotice(rm.Groups[1].Value, rm.Groups[2].Value, rm.Groups[3].Value);
                                    else // :FlashIrcServer.1 NOTICE AUTH :*** IRCwx 1.x auth type set.
                                        OnServerNotice(line.Substring(line.Substring(1).IndexOf(':') + 2));
                                    break;
                                case "WHISPER":
                                    r = new Regex(@"^:([^\40]+)!([^\40]+@[^\40]+)\40(?:WHISPER)\40:?[^\40]*\40[^\40]*\40:(.*)?$");
                                    rm = r.Match(line);
                                    if (rm.Success)
                                        OnWhisper(rm.Groups[1].Value, rm.Groups[2].Value, rm.Groups[3].Value);
                                    break;
                                case "PRIVMSG":
                                    r = new Regex(@"^:(%[^\40]+|[^\40!]+)(?:!([^\40]+@[^\40]+))?\40(?:PRIVMSG)\40[^\40]*(?:\40([^:]*))?\40:(.*)?$");
                                    rm = r.Match(line);
                                    if (rm.Success)
                                    {
                                        if (rm.Groups[3].Value == "")
                                            OnText(rm.Groups[1].Value, rm.Groups[2].Value, rm.Groups[4].Value);
                                        else
                                            OnQuery(rm.Groups[1].Value, rm.Groups[2].Value, rm.Groups[4].Value);
                                    }
                                    break;
                                case "KILL":
                                    r = new Regex(@"^:([^\40]+)!([^\40]+@[^\40]+)\40(?:KILL)\40([^\40]*)(?:\40:?(.*)?)?$");
                                    rm = r.Match(line);
                                    if (rm.Success)
                                        OnKill(rm.Groups[1].Value, rm.Groups[2].Value, rm.Groups[3].Value, rm.Groups[4].Value);
                                    break;
                                case "PROP":
                                    r = new Regex(@"^:([^\40]+)!([^\40]+@[^\40]+)\40(?:PROP)\40:?[^\40]*\40([^\40]*)\40:(.*)?$");
                                    rm = r.Match(line);
                                    if (rm.Success)
                                        OnProp(rm.Groups[1].Value, rm.Groups[2].Value, rm.Groups[3].Value, rm.Groups[4].Value);
                                    break;
                                case "NICK":
                                    r = new Regex(@"^:([^\40]+)?!([^\40]+@[^\40]+)\40(?:NICK)\40:?([^\40]*)(?:.*)?$");
                                    rm = r.Match(line);
                                    if (rm.Success)
                                        OnNickChange(rm.Groups[1].Value, rm.Groups[2].Value, rm.Groups[3].Value);
                                    break;
                                case "MODE":
                                    r = new Regex(@"^:([^\40]+)!([^\40]+@[^\40]+)\40(?:MODE)\40[^\40]*\40([^\40]*)(?:\40(.*))?$");
                                    rm = r.Match(line);
                                    if (rm.Success)
                                        OnMode(rm.Groups[1].Value, rm.Groups[2].Value, rm.Groups[3].Value, rm.Groups[4].Value);
                                    break;
                                case "DATA":
                                    // :FlashIrcServer.1 DATA Tewl PID :Tewl 211315
                                    // :FlashIrcServer.1 DATA Tewl PID :>Tewl -1
                                    if (words[3] == "PID")
                                    {
                                        if (words[5] != "-1")
                                            System.Diagnostics.Process.Start("http://www.buzzen.com/chat/redirect.php?mid=" + words[5]);
                                        else
                                            SetText("A profile does not exist for " + words[4].Substring(1), System.Windows.Media.Colors.Gray, false, false);
                                    }
                                    break;
                                default:
                                    SetText(line);
                                    break;
                            }
                        }
                    }
                }
            }
        }
        #endregion

        #region Encoding/Decoding/Encryption
        public string Encode(string text)
        {
            byte[] b = Encoding.Default.GetBytes(text);
            b = Encoding.Convert(Encoding.UTF8, Encoding.Unicode, b);
            return Encoding.Unicode.GetString(b);
        }

        public string Decode(string text)
        {
            byte[] b = Encoding.Unicode.GetBytes(text);
            b = Encoding.Convert(Encoding.Unicode, Encoding.UTF8, b);
            return Encoding.Default.GetString(b);
        }

        public string MD5Hash(string data)
        {
            System.Text.ASCIIEncoding ae = new System.Text.ASCIIEncoding();
            byte[] bytes = ae.GetBytes(data);
            System.Security.Cryptography.MD5 md5 = new System.Security.Cryptography.MD5CryptoServiceProvider();
            return BitConverter.ToString(md5.ComputeHash(bytes)).Replace("-", null).ToLower();
        }
        #endregion

        #region Font Style Methods
        private string GetStyle()
        {
            StringBuilder sb = new StringBuilder("[style ff:");
            sb.Append(InputFont.Name);
            sb.Append(";co:#");
            sb.Append(String.Format("{0:X2}",InputColor.R));
            sb.Append(String.Format("{0:X2}",InputColor.G));
            sb.Append(String.Format("{0:X2}",InputColor.B));
            if (InputFont.Bold)
                sb.Append(";b");
            if (InputFont.Italic)
                sb.Append(";i");
            sb.Append(";]");
            return sb.ToString();
        }

        private string MakeStyle(string fontname, System.Windows.Media.Color fontcolor, bool bold, bool italic)
        {
            StringBuilder sb = new StringBuilder("[style ff:");
            sb.Append(fontname);
            sb.Append(";co:#");
            sb.Append(String.Format("{0:X2}", fontcolor.R));
            sb.Append(String.Format("{0:X2}", fontcolor.G));
            sb.Append(String.Format("{0:X2}", fontcolor.B));
            if (bold)
                sb.Append(";b");
            if (italic)
                sb.Append(";i");
            sb.Append(";]");
            return sb.ToString();
        }

        private Font MakeFont(System.Drawing.FontFamily ff, bool bold, bool italic)
        {
            FontStyle fs;
            if ((bold) & (italic))
                fs = FontStyle.Bold | FontStyle.Italic;
            else if (bold)
                fs = FontStyle.Bold;
            else if (italic)
                fs = FontStyle.Italic;
            else
                fs = FontStyle.Regular;
            return new Font(ff, (float)rtbOutput.FontSize, fs, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
        }
        #endregion

        #region Socket Events/Methods
        private void OnConnect()
        {
            ReconnectAttempts = 0;
            AttemptReconnect = true;
            SetText("Connected!", System.Windows.Media.Colors.Green, false, false);
            SetTitle("DragonChat - Connection established. Attempting to join " + Channel);
            SendData("AUTHTYPE ircwx1");
            SendData("CLIENTMODE cd1");
            SendData("LOGINH " + Email + " " + MD5Hash(Password));
        }

        private void OnDisconnect()
        {
            try
            {
                SetText("Disconnected.", System.Windows.Media.Colors.Red, false, false);
                SetTitle("DragonChat");
                Topic = "";
                ToggleConnect(false);
                ClearNames();
                if (AttemptReconnect)
                {
                   Reconnect();
                }
            }
            catch (ObjectDisposedException)
            {
                // Ignore if program is closing.
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message,"OnDisconnect");
            }
        }

        private void OnDataArrival(string data)
        {
            int i = data.LastIndexOf("\r\n");
            if (i < 0)
                packet += data;
            else
            {
                packet += data.Substring(0, i);
                Parse(Encode(packet));
                packet = data.Substring(i + 2);
            }
        }

        public void SendData(string data)
        {
            if (Buzzen.IsConnected)
                Buzzen.SendData(data + "\r\n");
        }
        #endregion

        #region Menu Events
        private void connectToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            Connect();
        }

        private void disconnectToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            AttemptReconnect = false;
            Disconnect();
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Disconnect();
            Close();
        }

        private void supportForumToolStripMenuItem_Click(object sender, EventArgs e)
        {
            System.Diagnostics.Process.Start("http://forum.tewlzbox.com/viewforum.php?f=24");
        }

        private void aboutDragonChatToolStripMenuItem_Click(object sender, EventArgs e)
        {
            About af = new About();
            af.ShowDialog();
        }

        private void setupToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (setup == null)
            {
                setup = new Setup(this);
            }
            setup.ShowDialog();
        }

        private void whisperToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (snick != null)
            {
                if (Nickname != snick.Nickname)
                {
                    CreateWhisper(snick.Nickname);
                }
            }
        }

        private void viewProfileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (snick != null)
                SendData("PID " + snick.Nickname);
        }

        private void ignoreToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (snick != null)
            {
                System.Windows.Forms.ListViewItem lvi = GetItem(snick.Nickname);
                if (lvi != null)
                {
                    if (snick.Ignore)
                    {
                        snick.Ignore = false;
                    }
                    else
                    {
                        snick.Ignore = true;
                    }
                    ChangeIcon(lvi, snick.ImageIndex);
                }
            }
        }

        private void localTimeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (snick != null)
                SendData("PRIVMSG " + Channel + " " + snick.Nickname + " :TIME");
        }

        private void GiveOwnerMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
                SendData("MODE " + Channel + " +q " + snick.Nickname);
        }

        private void TakeOwnerMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
                SendData("MODE " + Channel + " -q " + snick.Nickname);
        }

        private void GiveHostMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
                SendData("MODE " + Channel + " +o " + snick.Nickname);
        }

        private void TakeHostMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
                SendData("MODE " + Channel + " -o " + snick.Nickname);
        }

        private void GiveHelpOpMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
                SendData("MODE " + Channel + " +h " + snick.Nickname);
        }

        private void TakeHelpOpMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
                SendData("MODE " + Channel + " -h " + snick.Nickname);
        }

        private void GiveVoiceMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
                SendData("MODE " + Channel + " +v " + snick.Nickname);
        }

        private void TakeVoiceMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
                SendData("MODE " + Channel + " -v " + snick.Nickname);
        }

        private void QuickKickMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
                SendData("KICK " + Channel + " " + snick.Nickname);
        }

        private void CustomKickMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
            {
                KickInputDialog kid = new KickInputDialog();

                if (kid.ShowDialog() == DialogResult.OK)
                {
                    string kickmsg = kid.Message;
                    SendData("KICK " + Channel + " " + snick.Nickname + " " + kickmsg);
                }
            }
        }

        private void RandomKickMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
                SendData("KICK " + Channel + " " + snick.Nickname + " " + RandomMessage());
        }

        private void QuickBanMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
            {
                if (snick.Address == "")
                    SendData("ACCESS " + Channel + " ADD DENY " + snick.Nickname + "!* 0 : DragonChat[" + snick.Nickname + "]");
                else
                    SendData("ACCESS " + Channel + " ADD DENY *!" + snick.Address + " 0 : DragonChat[" + snick.Nickname + "]");
                SendData("KICK " + Channel + " " + snick.Nickname);
            }
        }

        private void CustomBanMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
            {
                KickInputDialog kid = new KickInputDialog();

                if (kid.ShowDialog() == DialogResult.OK)
                {
                    string kickmsg = kid.Message;
                    if (snick.Address == "")
                        SendData("ACCESS " + Channel + " ADD DENY " + snick.Nickname + "!* 0 : DragonChat[" + snick.Nickname + "]");
                    else
                        SendData("ACCESS " + Channel + " ADD DENY *!" + snick.Address + " 0 : DragonChat[" + snick.Nickname + "]");
                    SendData("KICK " + Channel + " " + snick.Nickname + " " + kickmsg);
                }
            }
        }

        private void RandomBanMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
            {
                if (snick.Address == "")
                    SendData("ACCESS " + Channel + " ADD DENY " + snick.Nickname + "!* 0 : DragonChat[" + snick.Nickname + "]");
                else
                    SendData("ACCESS " + Channel + " ADD DENY *!" + snick.Address + " 0 : DragonChat[" + snick.Nickname + "]");
                SendData("KICK " + Channel + " " + snick.Nickname + " " + RandomMessage());
            }
        }

        private void ownerAccessMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
            {
                if (snick.Address == "")
                {
                    SetText("\t  Unable to find address for [style co:#FF0000;]" + snick.Nickname + "[/style].", System.Windows.Media.Colors.Gray, false, false);
                }
                else
                {
                    SendData("ACCESS " + Channel + " ADD OWNER *!" + snick.Address + " 0 : DragonChat[" + snick.Nickname + "]");
                }
            }
        }

        private void hostAccessMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
            {
                if (snick.Address == "")
                {
                    SetText("\t  Unable to find address for [style co:#FF0000;]" + snick.Nickname + "[/style].", System.Windows.Media.Colors.Gray, false, false);
                }
                else
                {
                    SendData("ACCESS " + Channel + " ADD HOST *!" + snick.Address + " 0 : DragonChat[" + snick.Nickname + "]");
                }
            }
        }

        private void voiceAccessMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
            {
                if (snick.Address == "")
                {
                    SetText("\t  Unable to find address for [style co:#FF0000;]" + snick.Nickname + "[/style].", System.Windows.Media.Colors.Gray, false, false);
                }
                else
                {
                    SendData("ACCESS " + Channel + " ADD VOICE *!" + snick.Address + " 0 : DragonChat[" + snick.Nickname + "]");
                }
            }
        }

        private void denyAccessMenu_Click(object sender, EventArgs e)
        {
            if (snick != null)
            {
                if (snick.Address == "")
                {
                    SetText("\t  Unable to find address for [style co:#FF0000;]" + snick.Nickname + "[/style].", System.Windows.Media.Colors.Gray, false, false);
                }
                else
                {
                    SendData("ACCESS " + Channel + " ADD DENY *!" + snick.Address + " 0 : DragonChat[" + snick.Nickname + "]");
                }
            }
        }

        private void AccessMangerMenu_Click(object sender, EventArgs e)
        {
            if (alist == null)
                alist = new AccessList(this);
            if (alist.IsDisposed)
                alist = new AccessList(this);
            SendData("ACCESS " + Channel + " list");
            alist.Show();
        }
        #endregion

        private string RandomMessage()
        {
            string kickmsg = string.Empty;
            if (!File.Exists(Application.StartupPath + "\\RandomKicks.txt"))
            {
                TextWriter tw = new StreamWriter(Application.StartupPath + "\\RandomKicks.txt");
                tw.WriteLine("This is a random kick message.");
                tw.Close();
            }
            TextReader tr = new StreamReader(Application.StartupPath + "\\RandomKicks.txt");
            int nooflines = 0;
            while (tr.ReadLine() != null)
            {
                nooflines++;
            }
            Random randnum = new Random();

            tr.Close();
            tr = new StreamReader(Application.StartupPath + "\\RandomKicks.txt");


            int rnd = randnum.Next(0, nooflines);
            int i = 0;
            string temp;
            while ((temp = tr.ReadLine()) != null)
            {

                if (i == rnd)
                {
                    kickmsg = temp;
                    break;
                }
                i++;

            }
            tr.Close();
            return kickmsg;
        }

        private bool IsNumeric(string value)
        {
            try
            {
                Convert.ToInt32(value);
                return true;
            }
            catch
            {
                return false;
            }

        }

        private string StripStyle(string value)
        {
            Regex r = new Regex(@"\[(?:/)?style(?:[^\]]+)?\]", RegexOptions.IgnoreCase);
            return r.Replace(value, "");
        }

        private string StripmIRC(string value)
        {
            Regex r = new Regex(@"(\u0003(\d(\d)?(,(\d(\d)?)?)?)?|\u001F|\u0002|\u000F|\u0016)");
            return r.Replace(value, "");
        }

        private void UpdateAccess()
        {
            if (this.InvokeRequired)
            {
                CmdCallback d = new CmdCallback(this.UpdateAccess);
                this.Invoke(d);
            }
            else
            {
                if (alist != null)
                {
                    if (!alist.IsDisposed)
                    {
                        alist.List = AccessEntries;
                        alist.UpdateList();
                    }
                }
            }
        }

        private void UpdateSettings()
        {
            if (this.InvokeRequired)
            {
                CmdCallback d = new CmdCallback(this.UpdateSettings);
                this.Invoke(d);
            }
            else
            {
                if (chanman != null)
                {
                    if (!chanman.IsDisposed)
                    {
                        chanman.Ownerkey = Ownerkey;
                        chanman.Hostkey = Hostkey;
                        chanman.Category = Category;
                        chanman.UpdateSettings();
                    }
                }
            }
        }

        public double GetTimeStamp()
        {
            return (DateTime.Now.ToUniversalTime().Ticks - 621355968000000000) / 10000000;
        }

        public DateTime GetDateTime(double timestamp)
        {
            DateTime dt = new DateTime(1970, 1, 1, 0, 0, 0, 0);
            dt = dt.AddSeconds(timestamp);
            return dt;
        }

        private void Client_Load(object sender, EventArgs e)
        {

        }

    }
}